import numpy as np


# Simple order
# set higher dipole moment states first if energy gap is lower than eps_energy
def get_order_states_list(states, eps_moment=0.1, eps_energy=0.05):

    import itertools
    order = np.arange(len(states))

    for subset in itertools.combinations(range(len(states)), 2):
        i, j = subset

        if np.abs(states[i]['total energy'] - states[j]['total energy']) < eps_energy:
            tmi = np.linalg.norm(states[i]['transition moment'])
            tmj = np.linalg.norm(states[j]['transition moment'])
            if tmi - tmj < eps_moment:
                order[i], order[j] = order[j], order[i]

    return order


def correct_order_list(list, order):

    if np.array(order).shape != np.array(list).T.shape:
        print(np.array(order).shape, np.array(list).T.shape)
        raise Exception('Error in correcting order (not same shape)')

    alist = np.array(list)

    try:
        ordered_list = []
        for l, o in zip(alist.T, order):
            # print('l', l, 'o', o)
            ordered_list.append(l[o])
    except:
        ordered_list = []
        for l, o in zip(alist, order):
            # print('l', l, 'o', o)
            ordered_list.append(list[o])

    return np.array(ordered_list).T.tolist()



__author__ = 'Abel Carreras'
from pyqchem.utils import search_bars


def _get_orbital_energies(orbitals_section):
    occupied = orbitals_section.find('-- Occupied --')
    virtual = orbitals_section.find('-- Virtual --')

    occupied_section = orbitals_section[occupied:virtual]
    virtual_section = orbitals_section[virtual:]

    occupied_energies = [float(energy) for energy in occupied_section.split()[3:]]
    virtual_energies = [float(energy) for energy in virtual_section.split()[3:]]

    return {'occupied': occupied_energies, 'virtual': virtual_energies}


def basic_parser_qchem(output):

    data_dict = {}

    # scf energy
    enum = output.find('Total energy in the final basis set')
    data_dict['scf_energy'] = float(output[enum:enum+100].split()[8])

    # Orbitals energy
    enum = output.find('Orbital Energies (a.u.)')
    bars = search_bars(output, from_position=enum, bar_type='----')
    orbitals_section = output[enum:bars[1]]

    alpha_mos = orbitals_section.find('Alpha MOs')
    beta_mos = orbitals_section.find('Beta MOs')

    if beta_mos > 0:
        alpha_energies = _get_orbital_energies(orbitals_section[alpha_mos:beta_mos])
        beta_energies = _get_orbital_energies(orbitals_section[beta_mos:])
    else:
        alpha_energies = _get_orbital_energies(orbitals_section[alpha_mos:])
        beta_energies = alpha_energies

    data_dict['orbital_energies'] = {'alpha': alpha_energies, 'beta': beta_energies, 'units': 'au'}

    # Mulliken Net Atomic Charges
    enum = output.find('Ground-State Mulliken Net Atomic Charges')
    bars = search_bars(output, from_position=enum, bar_type='----')
    mulliken_section = output[bars[0]:bars[1]]
    data_dict['mulliken_charges'] = [float(line.split()[2]) for line in mulliken_section.split('\n')[1:-1]]

    # Multipole Moments
    enum = output.find('Cartesian Multipole Moments')
    bars = search_bars(output, from_position=enum, bar_type='----')
    multipole_section = output[bars[0]: bars[1]]
    multipole_lines =  multipole_section.split('\n')[1:-1]

    multipole_dict = {}

    multipole_dict['charge'] = float(multipole_lines[1])
    multipole_dict['charge_units'] = 'ESU x 10^10'

    multipole_dict['dipole_moment'] = [float(val) for val in multipole_lines[3].split()[1::2]]
    multipole_dict['dipole_units'] = 'Debye'

    multipole_dict['quadrupole_moment'] = [float(val) for val in multipole_lines[6].split()[1::2]] + \
                                          [float(val) for val in multipole_lines[7].split()[1::2]]
    multipole_dict['quadrupole_units'] = 'Debye-Ang'

    multipole_dict['octopole_moment'] = [float(val) for val in multipole_lines[9].split()[1::2]] + \
                                        [float(val) for val in multipole_lines[10].split()[1::2]] + \
                                        [float(val) for val in multipole_lines[11].split()[1::2]] + \
                                        [float(val) for val in multipole_lines[12].split()[1::2]]
    multipole_dict['octopole_units'] = 'Debye-Ang^2'

    data_dict['multipole'] = multipole_dict

    return data_dict





from pyqchem.structure import Structure
import numpy as np


def basic_optimization(output, print_data=False):

    # Molecule
    n = output.find('$molecule')
    n2 = output[n:].find('$end')

    molecule_region = output[n:n+n2-1].replace('\t', ' ').split('\n')[1:]
    charge, multiplicity = [int(num) for num in molecule_region[0].split()]
    coordinates = np.array([np.array(line.split()[1:4], dtype=float) for line in molecule_region[1:]])
    symbols = [line.split()[0].capitalize() for line in molecule_region[1:]]
    n_atoms = len(coordinates)

    # Optimization
    n = output.find('**  OPTIMIZATION CONVERGED  **')
    ne = output[n-200:n].find('Final energy')

    final_energy = float(output[ne+n-200: n].split()[3])
    optimization_section = output[n:]
    coordinates_section = optimization_section.split('\n')
    coordinates_final = [line.split()[2:5] for line in coordinates_section[5:5+n_atoms]]

    optimized_molecule = Structure(coordinates=np.array(coordinates_final, dtype=float).tolist(),
                                   atomic_elements=symbols,
                                   charge=charge,
                                   multiplicity=multiplicity)

    return {'optimized_molecule': optimized_molecule,
            'energy': final_energy}





from wfnsympy import WfnSympy
import numpy as np
from pyqchem.utils import get_occupied_electrons, reorder_coefficients


def get_wf_symmetry(structure,
                    basis,
                    mo_coeff,
                    center=None,
                    orientation=(0., 0., 1.),
                    group='C2h'):

    """
    Simplified interface between pyqchem anf wfnsympy

    :param structure: molecular geometry (3xn) array
    :param basis: dictionary containing the basis
    :param mo_coeff: molecular orbitals coefficients
    :param center: center of the molecule
    :param orientation: unit vector perpendicular to the plane of the molecule
    :param group: point symmetry group
    :return molsym: wfnsympy object
    """

    alpha_mo_coeff = np.array(mo_coeff['alpha']).tolist()
    if 'beta' in mo_coeff:
        beta_mo_coeff = np.array(mo_coeff['beta']).tolist()
    else:
        beta_mo_coeff = None

    molsym = WfnSympy(coordinates=structure.get_coordinates(),
                      symbols=structure.get_atomic_elements(),
                      basis=basis,
                      center=center,
                      VAxis=orientation,
                      alpha_mo_coeff=alpha_mo_coeff,
                      beta_mo_coeff=beta_mo_coeff,
                      charge=structure.charge,
                      multiplicity=structure.multiplicity,
                      group=group)
    return molsym


def get_orbital_classification(fchk_data,
                               center=(0., 0., 0.),
                               orientation=(0., 0., 1.)):
    """
    Classify orbitals in sigma/pi in planar molecules

    :param fchk_data: dictionary containing fchk data
    :param orientation: unit vector perpendicular to the plane of the molecule
    :return: list of labels 'sigma'/'pi' according to the order of the orbitals
    """
    molsym = get_wf_symmetry(fchk_data['structure'],
                             fchk_data['basis'],
                             fchk_data['coefficients'],
                             center=center,
                             orientation=orientation)

    sh_index = molsym.SymLab.index('s_h')
    orbital_type_alpha = []
    for i, overlap in enumerate(molsym.mo_SOEVs_a[:, sh_index]):
        overlap = overlap / molsym.mo_SOEVs_a[i, molsym.SymLab.index('E')]  # normalize
        if overlap < 0:
            orbital_type_alpha.append(['pi', np.abs(overlap)])
        else:
            orbital_type_alpha.append(['sigma', np.abs(overlap)])

    if 'beta' in fchk_data['coefficients']:
        orbital_type_beta = []
        for i, overlap in enumerate(molsym.mo_SOEVs_b[:, sh_index]):
            if overlap < 0:
                orbital_type_beta.append(['pi', np.abs(overlap)])
            else:
                orbital_type_beta.append(['sigma', np.abs(overlap)])
        return orbital_type_alpha, orbital_type_beta
    else:
        return orbital_type_alpha




def get_state_symmetry(parsed_fchk,
                       rasci_states,
                       center=None,
                       orientation=(0, 0, 1),
                       group='C2h',
                       extra_print=False,
                       amplitude_cutoff=0.0
                       ):

    sym_states = {}
    for i, state in enumerate(rasci_states):

        structure = parsed_fchk['structure']
        total_orbitals = len(parsed_fchk['coefficients']['alpha'])

        if extra_print:
            print('HF/KS Ground state\n---------------------------')
            molsym = get_wf_symmetry(structure,
                                     parsed_fchk['basis'],
                                     parsed_fchk['coefficients'],
                                     center=center,
                                     orientation=orientation,
                                     group=group)

            molsym.print_alpha_mo_IRD()
            molsym.print_beta_mo_IRD()
            molsym.print_wf_mo_IRD()

            print('\nRASCI Exited state {}\n---------------------------'.format(i+1))

        # print(configurations)
        occupations_list = []
        for configuration in state['configurations']:
            if np.abs(configuration['amplitude']) < amplitude_cutoff:
                continue

            occupied_orbitals = get_occupied_electrons(configuration, structure)
            n_extra = total_orbitals - occupied_orbitals - len(configuration['alpha'])
            vector_alpha = [1] * occupied_orbitals + [int(c) for c in configuration['alpha']] + [0] * n_extra

            n_extra = total_orbitals - occupied_orbitals - len(configuration['beta'])
            vector_beta = [1] * occupied_orbitals + [int(c) for c in configuration['beta']] + [0] * n_extra

            occupations_list.append({'alpha': vector_alpha, 'beta': vector_beta})

            if extra_print:
                print(configuration['alpha'], configuration['beta'], configuration['amplitude'])



        state_symmetry_list = []
        for occupations in occupations_list:
            # print('occupations', occupations)

            reordered_coefficients = reorder_coefficients(occupations, parsed_fchk['coefficients'])

            molsym = get_wf_symmetry(structure,
                                     parsed_fchk['basis'],
                                     reordered_coefficients,
                                     center=center,
                                     orientation=orientation,
                                     group=group)

            if extra_print:
                molsym.print_alpha_mo_IRD()
                molsym.print_beta_mo_IRD()
                molsym.print_wf_mo_IRD()

            state_symmetry = {}
            for label, measure in zip(molsym.IRLab, molsym.wf_IRd):
                state_symmetry[label] = measure

            # return maximum symmetry and value
            if extra_print:
                print([molsym.IRLab[np.argmax(molsym.wf_IRd)], np.max(molsym.wf_IRd)])
            state_symmetry_list.append([molsym.IRLab[np.argmax(molsym.wf_IRd)], np.max(molsym.wf_IRd)])

        if extra_print:
            print(state_symmetry_list)

        # Make sure symmetry of all configurations is the same
        assert len(np.unique([a[0] for a in state_symmetry_list])) == 1

        symmetry_label = state_symmetry_list[0][0]
        average_measure = np.average([a[1] for a in state_symmetry_list])
        sym_states['state {}'.format(i+1)] = [symmetry_label, average_measure]
    return sym_states


if __name__ == '__main__':
    from pyqchem.qchem_core import get_output_from_qchem, create_qchem_input
    from pyqchem.structure import Structure
    from pyqchem.file_io import build_fchk
    from pyqchem.utils import set_zero_coefficients

    benzene_coordinates = [[ 0.00000,  1.40272, 0.00000],
                           [ 0.00000,  2.49029, 0.00000],
                           [-1.21479,  0.70136, 0.00000],
                           [-2.15666,  1.24515, 0.00000],
                           [-1.21479, -0.70136, 0.00000],
                           [-2.15666, -1.24515, 0.00000],
                           [ 0.00000, -1.40272, 0.00000],
                           [ 0.00000, -2.49029, 0.00000],
                           [ 1.21479, -0.70136, 0.00000],
                           [ 2.15666, -1.24515, 0.00000],
                           [ 1.21479,  0.70136, 0.00000],
                           [ 2.15666,  1.24515, 0.00000]]

    symbols = ['C','H','C','H','C','H','C','H','C','H','C','F']

    benzene_coordinates = np.array(benzene_coordinates)
    molecule = Structure(coordinates=benzene_coordinates,
                         atomic_elements=symbols,
                         charge=0)

    parameters = {'jobtype': 'sp',
                  'exchange': 'hf',
                  'basis': '6-31G'}

    qc_input = create_qchem_input(molecule, **parameters)

    _, _, fchk_data = get_output_from_qchem(qc_input, processors=4, force_recalculation=True,
                                            read_fchk=True)

    from pyqchem.file_io import build_fchk
    open('test.fchk', 'w').write(build_fchk(fchk_data))

    mo_coeff = set_zero_coefficients(fchk_data['basis'],
                                           fchk_data['coefficients'],
                                           range(6, 12))
    fchk_data['coefficients'] = mo_coeff

    structure = fchk_data['structure']
    print(structure.get_xyz())

    txt_fchk = build_fchk(fchk_data)
    open('test2.fchk', 'w').write(txt_fchk)

    z_dist = structure.get_coordinates()[0][2]

    orbital_type = get_orbital_classification(fchk_data, center=[0.0, 0.0, z_dist])

    print('  {:5} {:5} {:5}'.format('num', 'type', 'overlap'))
    for i, ot in enumerate(orbital_type):
        print('{:5}:  {:5} {:5.2f}'.format(i+1, ot[0], ot[1]))



class ParserError(Exception):
    def __init__(self, parser_name, message):
        self.parser_name = parser_name
        self.message = message

    def __str__(self):
        return 'Error found while parsing output using parser {}: {}'.format(self.parser_name, self.message)


class OutputError(Exception):
    def __init__(self, output, error_lines):
        self.full_output = output
        self.error_lines = error_lines

    def __str__(self):
        return 'Error in Q-Chem calculation:\n{}'.format(self.error_lines)

class QchemInputWarning(UserWarning):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


import re
import numpy as np


# parser for 4 states
def analyze_diabatic(output, print_data=False, state_threshold=0.2, n_mon=6):

    # Coordinates
    n = output.find('$molecule')
    n2 = output[n:].find('$end')
    molecule_region = output[n:n+n2-1].replace('\t', ' ').split('\n')[1:]
    coordinates = np.array([np.array(line.split()[1:4], dtype=float) for line in molecule_region[1:]])
    symbols = [line.split()[0].capitalize() for line in molecule_region[1:]]
    n_atoms = len(coordinates)

    # Diabatic states
    for i, line in enumerate(output.split('\n')):
        if 'adiabatic states' in line.lower():
            loc_diab = [int(num) for num in output.split('\n')[i+1].split()]

    if print_data:
        print('-------------------------------')
        print('Mulliken')

    # determine diabatic states type
    state_order = []
    for m in re.finditer('Mulliken analysis of TDA State', output):
        #print output[m.end():m.end()+800].split()
        data = output[m.end():m.end()+37*(n_atoms+3)].split()
        state_number = int(data[0])
        state_info = []
        for i in range(n_atoms):
            # print(data[7+i*4:7+i*4+4][1:4])
            dat = data[7+i*4:7+i*4+4][1:4]
            state_info.append([float(d) for d in dat])

        state_info = np.array(state_info)
        e_sum1 = np.sum(state_info[0:n_mon, 0])
        e_sum2 = np.sum(state_info[n_mon:n_atoms, 0])

        h_sum1 = np.sum(state_info[0:n_mon, 1])
        h_sum2 = np.sum(state_info[n_mon:n_atoms, 1])

        # print (state_info)
        # print ('e_sum', e_sum1, e_sum2)
        # print ('h_sum', h_sum1, h_sum2)

        label = ''
        if state_number in loc_diab:
            if ((np.abs(e_sum1) > 1-state_threshold and np.abs(e_sum2) < state_threshold) or
                (np.abs(e_sum1) < state_threshold and np.abs(e_sum2) > 1-state_threshold)) and \
                 np.abs(np.abs(e_sum1) - np.abs(h_sum1)) < state_threshold:

                if (e_sum1 - e_sum2) > 0:
                    label = '01'
                else:
                    label = '10'

                state_order.append(label)

            if ((np.abs(np.abs(e_sum1)) > 1-state_threshold and np.abs(h_sum1) < state_threshold) or
                (np.abs(np.abs(e_sum1)) < state_threshold and np.abs(h_sum1) > 1-state_threshold)) and \
                np.abs(np.abs(e_sum1) - np.abs(h_sum1)) > 1-state_threshold:

                if (e_sum1 - e_sum2) > 0:
                    label = 'CA'
                else:
                    label = 'AC'
                state_order.append(label)

        if print_data:
            print ('state: {}  {}'.format(state_number, label))

    # AdiabatH
    if len(state_order) < 4:
        raise Exception('Some states not found')

    if print_data:
        print ('-------------------------------')
    adiabatic_energies = {}
    for m in re.finditer('showmatrix adiabatH', output):
        #print output[m.end():m.end()+800].split()
        data = output[m.end():m.end()+30].split()
        indices = output[m.end()+1:m.end()+4].split(',')
        indices = [int(i) for i in indices]
        # print (indices)
        adiabat_h = float(data[2])
        # print (adiabat_h)
        if indices[0] == indices[1]:
            adiabatic_energies.update({'E_{}'.format(indices[0]+1): adiabat_h * 27.2114})

    if print_data:
        for item in sorted(adiabatic_energies):
            print('{:5} : {:10.5f}'.format(item, adiabatic_energies[item]))

        print('-------------------------------')

    if print_data:
        for item in sorted(adiabatic_energies):
            print('{:5} : {:10.5f}'.format(item, adiabatic_energies[item]))
        print('-------------------------------')

    diabatic_energies = {}
    for m in re.finditer('showmatrix diabatH', output):
        data = output[m.end():m.end()+30].split()
        indices = output[m.end()+1:m.end()+4].split(',')
        indices = [int(i) for i in indices]
        diabat_h = float(data[2])
        if indices[0] == indices[1]:
            # diabatic_energies.append(adiabat_h * 27.2114)
            if state_order[indices[0]] == '10':
                diabatic_energies.update({'E_10': diabat_h * 27.2114})
            if state_order[indices[0]] == '01':
                diabatic_energies.update({'E_01': diabat_h * 27.2114})
            if state_order[indices[0]] == 'AC':
                diabatic_energies.update({'E_AC': diabat_h * 27.2114})
            if state_order[indices[0]] == 'CA':
                diabatic_energies.update({'E_CA': diabat_h * 27.2114})

        diabatic_energies.update({'E_{}_{}'.format(*np.array(indices) + 1): diabat_h * 27.2114})

        # print ('test_indices:', indices, state_order, loc_diab)
        # print ('test_states:', [state_order[indices[0]], state_order[indices[1]]])

        if [state_order[indices[0]], state_order[indices[1]]] == ['10', '01']:
            diabatic_energies.update({'V_DC_1': diabat_h * 27.2114})

        if [state_order[indices[0]], state_order[indices[1]]] == ['01', '10']:
            diabatic_energies.update({'V_DC_2': diabat_h * 27.2114})

        if [state_order[indices[0]], state_order[indices[1]]] == ['AC', 'CA']:
            diabatic_energies.update({'V_CT_1': diabat_h * 27.2114})

        if [state_order[indices[0]], state_order[indices[1]]] == ['CA', 'AC']:
            diabatic_energies.update({'V_CT_2': diabat_h * 27.2114})

        if [state_order[indices[0]], state_order[indices[1]]] == ['10', 'CA']:
            diabatic_energies.update({'V_e_1': diabat_h * 27.2114})

        if [state_order[indices[0]], state_order[indices[1]]] == ['01', 'AC']:
            diabatic_energies.update({'V_e_2': diabat_h * 27.2114})

        if [state_order[indices[0]], state_order[indices[1]]] == ['10', 'AC']:
            diabatic_energies.update({'V_h_1': diabat_h * 27.2114})

        if [state_order[indices[0]], state_order[indices[1]]] == ['01', 'CA']:
            diabatic_energies.update({'V_h_2': diabat_h * 27.2114})

    diabatic_energies.update({'V_DC': [diabatic_energies['V_DC_1'],
                                       diabatic_energies['V_DC_2']]})
    del diabatic_energies['V_DC_1']
    del diabatic_energies['V_DC_2']

    diabatic_energies.update({'V_CT': [diabatic_energies['V_CT_1'],
                                       diabatic_energies['V_CT_2']]})
    del diabatic_energies['V_CT_1']
    del diabatic_energies['V_CT_2']

    # arrange diabatic energies in lists
    diabatic_energies.update({'E_LE': (diabatic_energies['E_10'] +
                                       diabatic_energies['E_01'])/2})

    diabatic_energies.update({'dE_LE': (diabatic_energies['E_10'] -
                                        diabatic_energies['E_01'])/2})

    diabatic_energies.update({'E_CT': np.average((diabatic_energies['E_AC'] +
                                                  diabatic_energies['E_CA'])/2)})

    diabatic_energies.update({'dE_CT': (diabatic_energies['E_AC'] -
                                        diabatic_energies['E_CA'])/2})

    diabatic_energies.update({'V_e': [diabatic_energies['V_e_1'],
                                      diabatic_energies['V_e_2']]})
    del diabatic_energies['V_e_1']
    del diabatic_energies['V_e_2']

    diabatic_energies.update({'V_h': [diabatic_energies['V_h_1'],
                                      diabatic_energies['V_h_2']]})
    del diabatic_energies['V_h_1']
    del diabatic_energies['V_h_2']


    if len(diabatic_energies) < 6:
        raise Exception('diabatic states not resolved!')

    # Showmatrix Decomposed diabatics
    for diabatic_type in ['One', 'K', 'J']:
        for m in re.finditer('showmatrix Decomposed_{}_diabatic'.format(diabatic_type), output):
            data = output[m.end():m.end()+30].split()
            indices = output[m.end()+1:m.end()+4].split(',')
            indices = [int(i) for i in indices]
            element_diabatic = float(data[2])
            if indices[0] == indices[1]:
                if state_order[indices[0]] == '10':
                    diabatic_energies.update({'{}_10'.format(diabatic_type): element_diabatic * 27.2114})
                if state_order[indices[0]] == '01':
                    diabatic_energies.update({'{}_01'.format(diabatic_type): element_diabatic * 27.2114})
                if state_order[indices[0]] == 'AC':
                    diabatic_energies.update({'{}_AC'.format(diabatic_type): element_diabatic * 27.2114})
                if state_order[indices[0]] == 'CA':
                    diabatic_energies.update({'{}_CA'.format(diabatic_type): element_diabatic * 27.2114})

            if [state_order[indices[0]], state_order[indices[1]]] == ['10', '01']:
                diabatic_energies.update({'{}_DC_1'.format(diabatic_type): element_diabatic * 27.2114})

            if [state_order[indices[0]], state_order[indices[1]]] == ['01', '10']:
                diabatic_energies.update({'{}_DC_2'.format(diabatic_type): element_diabatic * 27.2114})

            if [state_order[indices[0]], state_order[indices[1]]] == ['AC', 'CA']:
                diabatic_energies.update({'{}_CT_1'.format(diabatic_type): element_diabatic * 27.2114})

            if [state_order[indices[0]], state_order[indices[1]]] == ['CA', 'AC']:
                diabatic_energies.update({'{}_CT_2'.format(diabatic_type): element_diabatic * 27.2114})

            if [state_order[indices[0]], state_order[indices[1]]] == ['10', 'CA']:
                diabatic_energies.update({'{}_e_1'.format(diabatic_type): element_diabatic * 27.2114})

            if [state_order[indices[0]], state_order[indices[1]]] == ['01', 'AC']:
                diabatic_energies.update({'{}_e_2'.format(diabatic_type): element_diabatic * 27.2114})

            if [state_order[indices[0]], state_order[indices[1]]] == ['10', 'AC']:
                diabatic_energies.update({'{}_h_1'.format(diabatic_type): element_diabatic * 27.2114})

            if [state_order[indices[0]], state_order[indices[1]]] == ['01', 'CA']:
                diabatic_energies.update({'{}_h_2'.format(diabatic_type): element_diabatic * 27.2114})

        diabatic_energies.update({'{}_DC'.format(diabatic_type): [diabatic_energies['{}_DC_1'.format(diabatic_type)],
                                                                  diabatic_energies['{}_DC_2'.format(diabatic_type)]]})
        del diabatic_energies['{}_DC_1'.format(diabatic_type)]
        del diabatic_energies['{}_DC_2'.format(diabatic_type)]

        diabatic_energies.update({'{}_CT'.format(diabatic_type): [diabatic_energies['{}_CT_1'.format(diabatic_type)],
                                                                  diabatic_energies['{}_CT_2'.format(diabatic_type)]]})
        del diabatic_energies['{}_CT_1'.format(diabatic_type)]
        del diabatic_energies['{}_CT_2'.format(diabatic_type)]

        diabatic_energies.update({'{}_e'.format(diabatic_type): [diabatic_energies['{}_e_1'.format(diabatic_type)],
                                                                 diabatic_energies['{}_e_2'.format(diabatic_type)]]})
        del diabatic_energies['{}_e_1'.format(diabatic_type)]
        del diabatic_energies['{}_e_2'.format(diabatic_type)]

        diabatic_energies.update({'{}_h'.format(diabatic_type): [diabatic_energies['{}_h_1'.format(diabatic_type)],
                                                                 diabatic_energies['{}_h_2'.format(diabatic_type)]]})
        del diabatic_energies['{}_h_1'.format(diabatic_type)]
        del diabatic_energies['{}_h_2'.format(diabatic_type)]


    if print_data:
        for item in sorted(diabatic_energies):
            try:
                print('{:5} : {:10.5f}'.format(item, diabatic_energies[item]))
            except:
                print('{:5} : '.format(item) + ' '.join(['{:10.5f}'.format(num) for num in diabatic_energies[item]]))

        print ('-------------------------------')

    coefficients = {}
    for m in re.finditer('showmatrix final adiabatic -> diabatic RotMatrix', output):
        data = output[m.end():m.end()+30].split()
        # print data
        indices = output[m.end()+1:m.end()+4].split(',')
        indices = [int(i) for i in indices]
        adiabat_f = float(data[2])

        for e_index in range(4):
            if [indices[0], state_order[indices[1]]] == [e_index, '10']:
                coefficients.update({'S{}_C10'.format(e_index+1): adiabat_f})
            if [indices[0], state_order[indices[1]]] == [e_index, '01']:
                coefficients.update({'S{}_C01'.format(e_index+1): adiabat_f})
            if [indices[0], state_order[indices[1]]] == [e_index, 'CA']:
                coefficients.update({'S{}_CCA'.format(e_index+1): adiabat_f})
            if [indices[0], state_order[indices[1]]] == [e_index, 'AC']:
                coefficients.update({'S{}_CAC'.format(e_index+1): adiabat_f})

            coefficients.update({'S{}_C{}'.format(*np.array(indices) + 1): adiabat_f})

    coefficients.update({'S_10': [coefficients.pop('S{}_C10'.format(i+1)) for i in range(4)]})
    coefficients.update({'S_01': [coefficients.pop('S{}_C01'.format(i+1)) for i in range(4)]})
    coefficients.update({'S_CA': [coefficients.pop('S{}_CCA'.format(i+1)) for i in range(4)]})
    coefficients.update({'S_AC': [coefficients.pop('S{}_CAC'.format(i+1)) for i in range(4)]})

    if print_data:
        for item in sorted(coefficients):
            try:
                print('{:5} : {:10.5f}'.format(item, coefficients[item]))
            except:
                print('{:5} : '.format(item) + ' '.join(['{:10.5f}'.format(num) for num in coefficients[item]]))

    diabatic_contributions = {}
    w_dc = []
    w_ct = []
    w_e = []
    w_h = []
    for i in range(4):
        factor = [coefficients['S_10'][i] * coefficients['S_01'][i],
                  coefficients['S_01'][i] * coefficients['S_10'][i]]
        w_dc.append(np.average([diabatic_energies['V_DC'][0] * np.sign(factor[0]),
                                diabatic_energies['V_DC'][1] * np.sign(factor[1])]))

        factor = [coefficients['S_AC'][i] * coefficients['S_CA'][i],
                  coefficients['S_CA'][i] * coefficients['S_AC'][i]]
        w_ct.append(np.average([diabatic_energies['V_CT'][0] * np.sign(factor[0]),
                                diabatic_energies['V_CT'][1] * np.sign(factor[1])]))

        factor = [coefficients['S_10'][i] * coefficients['S_AC'][i],
                  coefficients['S_01'][i] * coefficients['S_CA'][i]]
        w_h.append(np.average([diabatic_energies['V_h'][0] * np.sign(factor[0]),
                               diabatic_energies['V_h'][1] * np.sign(factor[1])]))

        factor = [coefficients['S_10'][i] * coefficients['S_CA'][i],
                  coefficients['S_01'][i] * coefficients['S_AC'][i]]
        w_e.append(np.average([diabatic_energies['V_e'][0] * np.sign(factor[0]),
                               diabatic_energies['V_e'][1] * np.sign(factor[1])]))

    diabatic_contributions.update({'W_DC': w_dc})
    diabatic_contributions.update({'W_CT': w_ct})
    diabatic_contributions.update({'W_h': w_h})
    diabatic_contributions.update({'W_e': w_e})

    if print_data:
        print ('-------------------------------')
        for item in ['W_DC', 'W_CT', 'W_h', 'W_e']:
            print('{:5} : {:10.5f} {:10.5f} {:10.5f} {:10.5f}'.format(
                *[item] + diabatic_contributions[item] ))

    for diabatic_type in ['One', 'K', 'J']:
        element_dc = []
        element_ct = []
        element_e = []
        element_h = []
        for i in range(4):
            factor = [coefficients['S_10'][i] * coefficients['S_01'][i],
                      coefficients['S_01'][i] * coefficients['S_10'][i]]
            element_dc.append(np.average([diabatic_energies['{}_DC'.format(diabatic_type)][0] * np.sign(factor[0]),
                                          diabatic_energies['{}_DC'.format(diabatic_type)][1] * np.sign(factor[1])]))

            factor = [coefficients['S_AC'][i] * coefficients['S_CA'][i],
                      coefficients['S_CA'][i] * coefficients['S_AC'][i]]
            element_ct.append(np.average([diabatic_energies['{}_CT'.format(diabatic_type)][0] * np.sign(factor[0]),
                                          diabatic_energies['{}_CT'.format(diabatic_type)][1] * np.sign(factor[1])]))

            factor = [coefficients['S_10'][i] * coefficients['S_AC'][i],
                      coefficients['S_01'][i] * coefficients['S_CA'][i]]
            element_h.append(np.average([diabatic_energies['{}_h'.format(diabatic_type)][0] * np.sign(factor[0]),
                                         diabatic_energies['{}_h'.format(diabatic_type)][1] * np.sign(factor[1])]))

            factor = [coefficients['S_10'][i] * coefficients['S_CA'][i],
                      coefficients['S_01'][i] * coefficients['S_AC'][i]]
            element_e.append(np.average([diabatic_energies['{}_e'.format(diabatic_type)][0] * np.sign(factor[0]),
                                         diabatic_energies['{}_e'.format(diabatic_type)][1] * np.sign(factor[1])]))

        diabatic_contributions.update({'W_{}_DC'.format(diabatic_type): element_dc})
        diabatic_contributions.update({'W_{}_CT'.format(diabatic_type): element_ct})
        diabatic_contributions.update({'W_{}_h'.format(diabatic_type): element_h})
        diabatic_contributions.update({'W_{}_e'.format(diabatic_type): element_e})

        if print_data:
            print('-------------------------------')
            for item in ['W_{}_DC'.format(diabatic_type),
                         'W_{}_CT'.format(diabatic_type),
                         'W_{}_h'.format(diabatic_type),
                         'W_{}_e'.format(diabatic_type)]:
                print('{:5} : {:10.5f} {:10.5f} {:10.5f} {:10.5f}'.format(
                    *[item] + diabatic_contributions[item]))

    l2 = []
    d2le = []
    d2ct = []
    for i in range(4):
        l2.append(coefficients['S_AC'][i]**2 + coefficients['S_CA'][i]**2)
        d2le.append(coefficients['S_10'][i]**2 - coefficients['S_01'][i]**2)
        d2ct.append(coefficients['S_AC'][i]**2 - coefficients['S_CA'][i]**2)

    asymmetries = {}
    asymmetries.update({'lambda2': l2})
    asymmetries.update({'delta2_le': d2le})
    asymmetries.update({'delta2_ct': d2ct})


    if print_data:
        print ('-------------------------------')
        for i, _l in enumerate(l2):
            print ('{}: lambda^2: {:10.5f}  c^2_le: {:10.5f}  c^2_ct: {:10.5f}'.format(i+1, _l, d2le[i], d2ct[i]))

    omega_dc = []
    for i in range(4):
        omega_dc.append(np.sqrt((1 - l2[i]) ** 2 - d2le[i] ** 2) * diabatic_contributions['W_DC'][i])

    omega_ct = []
    for i in range(4):
        omega_ct.append(np.sqrt(l2[i] ** 2 - d2ct[i] ** 2) * diabatic_contributions['W_CT'][i])

    omega_e = []
    for i in range(4):
        omega_e.append(2 * (coefficients['S_10'][i] * coefficients['S_CA'][i] * diabatic_energies['V_e'][0] +
                            coefficients['S_01'][i] * coefficients['S_AC'][i] * diabatic_energies['V_e'][1]))

    omega_h = []
    for i in range(4):
        omega_h.append(2 * (coefficients['S_10'][i] * coefficients['S_AC'][i] * diabatic_energies['V_h'][0] +
                            coefficients['S_01'][i] * coefficients['S_CA'][i] * diabatic_energies['V_h'][1]))

    diabatic_contributions.update({'Omega_dc': omega_dc})
    diabatic_contributions.update({'Omega_ct': omega_ct})

    diabatic_contributions.update({'Omega_e': omega_e})
    diabatic_contributions.update({'Omega_h': omega_h})

    for diabatic_type in ['One', 'K', 'J']:

        omega_e_dc = []
        for i in range(4):
            omega_e_dc.append(np.sqrt((1 - l2[i]) ** 2 - d2le[i] ** 2) *
                            diabatic_contributions['W_{}_DC'.format(diabatic_type)][i])

        omega_e_ct = []
        for i in range(4):
            omega_e_ct.append(np.sqrt(l2[i] ** 2 - d2ct[i] ** 2) *
                            diabatic_contributions['W_{}_CT'.format(diabatic_type)][i])

        omega_e_e = []
        for i in range(4):
            omega_e_e.append(2 * (coefficients['S_10'][i] * coefficients['S_CA'][i] *
                                  diabatic_energies['{}_e'.format(diabatic_type)][0] +
                                  coefficients['S_01'][i] * coefficients['S_AC'][i] *
                                  diabatic_energies['{}_e'.format(diabatic_type)][1]))
        omega_e_h = []
        for i in range(4):
            omega_e_h.append(2 * (coefficients['S_10'][i] * coefficients['S_AC'][i] *
                                  diabatic_energies['{}_h'.format(diabatic_type)][0] +
                                  coefficients['S_01'][i] * coefficients['S_CA'][i] *
                                  diabatic_energies['{}_h'.format(diabatic_type)][1]))

        diabatic_contributions.update({'Omega_{}_dc'.format(diabatic_type): omega_e_dc})
        diabatic_contributions.update({'Omega_{}_ct'.format(diabatic_type): omega_e_ct})

        diabatic_contributions.update({'Omega_{}_e'.format(diabatic_type): omega_e_e})
        diabatic_contributions.update({'Omega_{}_h'.format(diabatic_type): omega_e_h})

    reordering = []
    for m in re.finditer('Reordering necessary!', output):
        reordering.append([int(num) for num in output[m.end():m.end()+7].split('->')])

    return {'loc_diab': loc_diab,
            'adiabatic_energies': adiabatic_energies,
            'diabatic_energies': diabatic_energies,
            'diabatic_contributions': diabatic_contributions,
            'coefficients': coefficients,
            'asymmetries': asymmetries,
            'reordering': reordering}



import re
import numpy as np


# parser for 4 states
def analyze_diabatic(output, print_data=False, state_threshold=0.2, n_mon=6):

    # Coordinates
    n = output.find('$molecule')
    n2 = output[n:].find('$end')
    molecule_region = output[n:n+n2-1].replace('\t', ' ').split('\n')[1:]
    coordinates = np.array([ np.array(line.split()[1:4], dtype=float) for line in molecule_region[1:]])
    symbols = [line.split()[0].capitalize() for line in molecule_region[1:]]
    n_atoms = len(coordinates)

    # Diabatic states
    for i, line in enumerate(output.split('\n')):
        if 'adiabatic states' in line.lower():
            print('x')
            loc_diab = [int(num) for num in output.split('\n')[i+1].split()]

    if print_data:
        print('-------------------------------')
        print('Mulliken')

    state_order = []
    for m in re.finditer('Mulliken analysis of TDA State', output):
        #print output[m.end():m.end()+800].split()
        data = output[m.end():m.end()+37*(n_atoms+3)].split()
        state_number = int(data[0])
        state_info = []
        for i in range(n_atoms):
            # print(data[7+i*4:7+i*4+4][1:4])
            dat = data[7+i*4:7+i*4+4][1:4]
            state_info.append([float(d) for d in dat])

        state_info = np.array(state_info)
        e_sum1 = np.sum(state_info[0:n_mon, 0])
        e_sum2 = np.sum(state_info[n_mon:n_atoms, 0])

        h_sum1 = np.sum(state_info[0:n_mon, 1])
        h_sum2 = np.sum(state_info[n_mon:n_atoms, 1])

        # print (state_info)
        # print ('e_sum', e_sum1, e_sum2)
        # print ('h_sum', h_sum1, h_sum2)

        label = ''
        if state_number in loc_diab:
            if ((np.abs(e_sum1) > 1-state_threshold and np.abs(e_sum2) < state_threshold) or
                (np.abs(e_sum1) < state_threshold and np.abs(e_sum2) > 1-state_threshold)) and \
                 np.abs(np.abs(e_sum1) - np.abs(h_sum1)) < state_threshold:

                if (e_sum1 - e_sum2) > 0:
                    label = '01'
                else:
                    label = '10'

                state_order.append(label)

            if ((np.abs(np.abs(e_sum1)) > 1-state_threshold and np.abs(h_sum1) < state_threshold) or
                (np.abs(np.abs(e_sum1)) < state_threshold and np.abs(h_sum1) > 1-state_threshold)) and \
                np.abs(np.abs(e_sum1) - np.abs(h_sum1)) > 1-state_threshold:

                if (e_sum1 - e_sum2) > 0:
                    label = 'CA'
                else:
                    label = 'AC'
                state_order.append(label)

        if print_data:
            print ('state: {}  {}'.format(state_number, label))

    # AdiabatH
    if len(state_order) < 4:
        raise Exception('Some states not found')

    if print_data:
        print ('-------------------------------')
    adiabatic_energies = {}
    for m in re.finditer('showmatrix adiabatH', output):
        #print output[m.end():m.end()+800].split()
        data = output[m.end():m.end()+30].split()
        indices = output[m.end()+1:m.end()+4].split(',')
        indices = [int(i) for i in indices]
        # print (indices)
        adiabat_h = float(data[2])
        # print (adiabat_h)
        if indices[0] == indices[1]:
            adiabatic_energies.update({'E_{}'.format(indices[0]+1): adiabat_h * 27.2114})

    if print_data:
        for item in sorted(adiabatic_energies):
            print('{:5} : {:10.5f}'.format(item, adiabatic_energies[item]))

        print ('-------------------------------')

    if print_data:
        for item in sorted(adiabatic_energies):
            print('{:5} : {:10.5f}'.format(item, adiabatic_energies[item]))
        print ('-------------------------------')

    diabatic_energies = {}
    for m in re.finditer('showmatrix diabatH', output):
        data = output[m.end():m.end()+30].split()
        indices = output[m.end()+1:m.end()+4].split(',')
        indices = [int(i) for i in indices]
        adiabat_h = float(data[2])
        if indices[0] == indices[1]:
            # diabatic_energies.append(adiabat_h * 27.2114)
            if indices[0] == 0:
                diabatic_energies.update({'E_LE_1': adiabat_h * 27.2114})
            if indices[0] == 1:
                diabatic_energies.update({'E_LE_2': adiabat_h * 27.2114})
            if indices[0] == 2:
                diabatic_energies.update({'E_CT_1': adiabat_h * 27.2114})
            if indices[0] == 3:
                diabatic_energies.update({'E_CT_2': adiabat_h * 27.2114})

        diabatic_energies.update({'E_{}_{}'.format(*np.array(indices) + 1): adiabat_h * 27.2114})

        # print ('test_indices:', indices, state_order, loc_diab)
        # print ('test_states:', [state_order[indices[0]], state_order[indices[1]]])

        if [state_order[indices[0]], state_order[indices[1]]] == ['01', '10']:
            diabatic_energies.update({'V_DC': adiabat_h * 27.2114})

        if [state_order[indices[0]], state_order[indices[1]]] == ['AC', 'CA']:
            diabatic_energies.update({'V_CT': adiabat_h * 27.2114})

        if [state_order[indices[0]], state_order[indices[1]]] == ['10', 'CA']:
            diabatic_energies.update({'V_e_1': adiabat_h * 27.2114})

        if [state_order[indices[0]], state_order[indices[1]]] == ['01', 'AC']:
            diabatic_energies.update({'V_e_2': adiabat_h * 27.2114})

        if [state_order[indices[0]], state_order[indices[1]]] == ['10', 'AC']:
            diabatic_energies.update({'V_h_1': adiabat_h * 27.2114})

        if [state_order[indices[0]], state_order[indices[1]]] == ['01', 'CA']:
            diabatic_energies.update({'V_h_2': adiabat_h * 27.2114})

    # arrange diabatic energies in lists
    diabatic_energies.update({'E_LE': [diabatic_energies['E_LE_1'],
                                       diabatic_energies['E_LE_2']]})
    del diabatic_energies['E_LE_1']
    del diabatic_energies['E_LE_2']

    diabatic_energies.update({'E_CT': [diabatic_energies['E_CT_1'],
                                       diabatic_energies['E_CT_2']]})
    del diabatic_energies['E_CT_1']
    del diabatic_energies['E_CT_2']

    diabatic_energies.update({'V_e': [diabatic_energies['V_e_1'],
                                      diabatic_energies['V_e_2']]})
    del diabatic_energies['V_e_1']
    del diabatic_energies['V_e_2']

    diabatic_energies.update({'V_h': [diabatic_energies['V_h_1'],
                                      diabatic_energies['V_h_2']]})
    del diabatic_energies['V_h_1']
    del diabatic_energies['V_h_2']

    if print_data:
        for item in sorted(diabatic_energies):
            try:
                print('{:5} : {:10.5f}'.format(item, diabatic_energies[item]))
            except:
                print('{:5} : '.format(item) + ' '.join(['{:10.5f}'.format(num) for num in diabatic_energies[item]]))

        print ('-------------------------------')

    if len(diabatic_energies) < 6:
        raise Exception('diabatic states not resolved!')

    coefficients = {}
    for m in re.finditer('showmatrix final adiabatic -> diabatic RotMatrix', output):
        data = output[m.end():m.end()+30].split()
        # print data
        indices = output[m.end()+1:m.end()+4].split(',')
        indices = [int(i) for i in indices]
        adiabat_f = float(data[2])

        for e_index in range(4):
            if [indices[0], state_order[indices[1]]] == [e_index, '10']:
                coefficients.update({'S{}_C10'.format(e_index+1): adiabat_f})
            if [indices[0], state_order[indices[1]]] == [e_index, '01']:
                coefficients.update({'S{}_C01'.format(e_index+1): adiabat_f})
            if [indices[0], state_order[indices[1]]] == [e_index, 'CA']:
                coefficients.update({'S{}_CCA'.format(e_index+1): adiabat_f})
            if [indices[0], state_order[indices[1]]] == [e_index, 'AC']:
                coefficients.update({'S{}_CAC'.format(e_index+1): adiabat_f})

            coefficients.update({'S{}_C{}'.format(*np.array(indices) + 1): adiabat_f})

    coefficients.update({'S_10': [coefficients.pop('S{}_C10'.format(i+1)) for i in range(4)]})
    coefficients.update({'S_01': [coefficients.pop('S{}_C01'.format(i+1)) for i in range(4)]})
    coefficients.update({'S_CA': [coefficients.pop('S{}_CCA'.format(i+1)) for i in range(4)]})
    coefficients.update({'S_AC': [coefficients.pop('S{}_CAC'.format(i+1)) for i in range(4)]})

    if print_data:
        for item in sorted(coefficients):
            try:
                print('{:5} : {:10.5f}'.format(item, coefficients[item]))
            except:
                print('{:5} : '.format(item) + ' '.join(['{:10.5f}'.format(num) for num in coefficients[item]]))

    diabatic_contributions = {}
    w_dc = []
    w_ct = []
    w_e = []
    w_h = []
    for i in range(4):
        factor = coefficients['S_10'][i] * coefficients['S_01'][i]
        w_dc.append(diabatic_energies['V_DC'] * np.sign(factor))

        factor = coefficients['S_CA'][i] * coefficients['S_AC'][i]
        w_ct.append(diabatic_energies['V_CT'] * np.sign(factor))

        factor = [coefficients['S_10'][i] * coefficients['S_AC'][i],
                  coefficients['S_01'][i] * coefficients['S_CA'][i]]
        w_h.append(np.average([diabatic_energies['V_h'][0] * np.sign(factor[0]),
                               diabatic_energies['V_h'][1] * np.sign(factor[1])]))

        factor = [coefficients['S_10'][i] * coefficients['S_CA'][i],
                  coefficients['S_01'][i] * coefficients['S_AC'][i]]
        w_e.append(np.average([diabatic_energies['V_e'][0] * np.sign(factor[0]),
                               diabatic_energies['V_e'][1] * np.sign(factor[1])]))

    diabatic_contributions.update({'W_DC': w_dc})
    diabatic_contributions.update({'W_CT': w_ct})
    diabatic_contributions.update({'W_h': w_h})
    diabatic_contributions.update({'W_e': w_e})

    if True:
        print ('-------------------------------')
        for item in ['W_DC', 'W_CT', 'W_h', 'W_e']:
            print('{:5} : {:10.5f} {:10.5f} {:10.5f} {:10.5f}'.format(
                *[item] + diabatic_contributions[item] ))

    l = []
    for i in range(4):
        l.append(np.sqrt(2) * np.abs(coefficients['S_AC'][i]))


    if print_data:
        print ('-------------------------------')
        for i, _l in enumerate(l):
            print ('lambda/lambda^2 {}: {:10.5f} / {:10.5f}'.format(i+1, _l, _l**2))

    diabatic_contributions.update({'Omega_h': [2 * _l * np.sqrt(1 - _l**2) * _wh
                                               for _l, _wh in zip(l, diabatic_contributions['W_h'])]})
    diabatic_contributions.update({'Omega_e': [2 * _l * np.sqrt(1 - _l**2) * _we
                                               for _l, _we in zip(l, diabatic_contributions['W_e'])]})

    reordering = []
    for m in re.finditer('Reordering necessary!', output):
        reordering.append([int(num) for num in output[m.end():m.end()+7].split('->')])

    return {'loc_diab': loc_diab,
            'adiabatic_energies': adiabatic_energies,
            'diabatic_energies': diabatic_energies,
            'diabatic_contributions': diabatic_contributions,
            'coefficients': coefficients,
            'lambda': l,
            'reordering': reordering}




from pyqchem.structure import atom_data, Structure

import numpy as np
angstrom_to_bohr = 1/0.529177249

def get_array_txt(label, type, array, row_size=5):

    formats = {'R': '15.8e',
               'I': '11'}

    n_elements = len(array)
    rows = int(np.ceil(n_elements/row_size))

    txt_fchk = '{:40}   {}   N=       {:5}\n'.format(label, type, n_elements)

    # print(rows)
    for i in range(rows):
        if (i+1)*row_size > n_elements:
            txt_fchk += (' {:{fmt}}'* (n_elements - i*row_size) + '\n').format(*array[i * row_size:n_elements],
                                                                               fmt=formats[type])
        else:
            txt_fchk += (' {:{fmt}}'* row_size  + '\n').format(*array[i * row_size: (i+1)*row_size],
                                                               fmt=formats[type])

    return txt_fchk


def build_fchk(parsed_data):

    structure = parsed_data['structure']
    basis = parsed_data['basis']
    alpha_mo_coeff = parsed_data['coefficients']['alpha']
    alpha_mo_energies = parsed_data['mo_energies']['alpha']

    #overlap = parsed_data['overlap']
    #coor_shell = parsed_data['coor_shell']
    #core_hamiltonian = parsed_data['core_hamiltonian']
    #scf_density = parsed_data['scf_density']

    number_of_basis_functions = len(alpha_mo_coeff)
    number_of_electrons = np.sum(structure.get_atomic_numbers()) - structure.charge
    if structure.multiplicity > 1:
        # raise Exception('1> multiplicity not yet implemented')
        pass
    alpha_electrons = number_of_electrons // 2
    beta_electrons = number_of_electrons // 2
    #print(alpha_electrons)
    #print(number_of_electrons)

    alpha_mo_coeff = np.array(alpha_mo_coeff).flatten().tolist()

    if 'beta' in parsed_data['coefficients']:
        beta_mo_coeff = parsed_data['coefficients']['beta']
        beta_mo_coeff = np.array(beta_mo_coeff).flatten().tolist()
        beta_mo_energies = parsed_data['mo_energies']['beta']

    shell_type_list = {'s':  {'type':  0, 'angular_momentum': 0},
                       'p':  {'type':  1, 'angular_momentum': 1},
                       'd':  {'type':  2, 'angular_momentum': 2},
                       'f':  {'type':  3, 'angular_momentum': 3},
                       'sp': {'type': -1, 'angular_momentum': 1},  # hybrid
                       'dc': {'type': -2, 'angular_momentum': 2},  # pure
                       'fc': {'type': -3, 'angular_momentum': 3}}  # pure

    shell_type = []
    p_exponents = []
    c_coefficients = []
    p_c_coefficients = []
    n_primitives = []
    atom_map = []

    largest_degree_of_contraction = 0
    highest_angular_momentum = 0
    number_of_contracted_shells = 0

    for i, atoms in enumerate(basis['atoms']):
        for shell in atoms['shells']:
            number_of_contracted_shells += 1
            st = shell['shell_type']
            shell_type.append(shell_type_list[st]['type'])
            n_primitives.append(len(shell['p_exponents']))
            atom_map.append(i+1)
            if highest_angular_momentum < shell_type_list[st]['angular_momentum']:
                highest_angular_momentum = shell_type_list[st]['angular_momentum']

            if len(shell['con_coefficients']) > largest_degree_of_contraction:
                    largest_degree_of_contraction = len(shell['con_coefficients'])

            for p in shell['p_exponents']:
                p_exponents.append(p)
            for c in shell['con_coefficients']:
                c_coefficients.append(c)
            for pc in shell['p_con_coefficients']:
                p_c_coefficients.append(pc)

    coordinates_list = angstrom_to_bohr*np.array(structure.get_coordinates()).flatten()

    txt_fchk = '{}\n'.format('filename')
    txt_fchk += 'SP        R                             {}\n'.format('6-31G')
    txt_fchk += 'Number of atoms                            I               {}\n'.format(structure.get_number_of_atoms())
    txt_fchk += 'Charge                                     I               {}\n'.format(structure.charge)
    txt_fchk += 'Multiplicity                               I               {}\n'.format(structure.multiplicity)
    txt_fchk += 'Number of electrons                        I               {}\n'.format(number_of_electrons)
    txt_fchk += 'Number of alpha electrons                  I               {}\n'.format(alpha_electrons)
    txt_fchk += 'Number of beta electrons                   I               {}\n'.format(beta_electrons)

    txt_fchk += get_array_txt('Atomic numbers', 'I', structure.get_atomic_numbers(), row_size=6)
    txt_fchk += get_array_txt('Current cartesian coordinates', 'R', coordinates_list)
    txt_fchk += get_array_txt('Nuclear charges', 'R', structure.get_atomic_numbers())

    txt_fchk += 'Number of basis functions                  I               {}\n'.format(number_of_basis_functions)
    txt_fchk += 'Number of contracted shells                I               {}\n'.format(number_of_contracted_shells)
    txt_fchk += 'Number of primitive shells                 I               {}\n'.format(np.sum(n_primitives))
    txt_fchk += 'Highest angular momentum                   I               {}\n'.format(highest_angular_momentum)
    txt_fchk += 'Largest degree of contraction              I               {}\n'.format(largest_degree_of_contraction)

    txt_fchk += get_array_txt('Shell types', 'I', shell_type, row_size=6)
    txt_fchk += get_array_txt('Number of primitives per shell', 'I', n_primitives, row_size=6)
    txt_fchk += get_array_txt('Shell to atom map', 'I', atom_map, row_size=6)
    txt_fchk += get_array_txt('Primitive exponents', 'R', p_exponents)
    txt_fchk += get_array_txt('Contraction coefficients', 'R', c_coefficients)
    txt_fchk += get_array_txt('P(S=P) Contraction coefficients', 'R', p_c_coefficients)
    # txt_fchk += get_array_txt('Coordinates of each shell', 'R', coor_shell) #
    # txt_fchk += get_array_txt('Overlap Matrix', 'R', overlap)
    #txt_fchk += get_array_txt('Core Hamiltonian Matrix', 'R', core_hamiltonian)
    txt_fchk += get_array_txt('Alpha Orbital Energies', 'R', alpha_mo_energies)
    # txt_fchk += get_array_txt('Beta Orbital Energies', 'R', beta_mo_energies)
    # txt_fchk += get_array_txt('Total SCF Density', 'R', scf_density)
    txt_fchk += get_array_txt('Alpha MO coefficients', 'R', alpha_mo_coeff)
    # txt_fchk += get_array_txt('Beta MO coefficients', 'R', beta_mo_coeff)

    return txt_fchk

if __name__ == '__main__':
    from pyqchem.parsers.parser_fchk import parser_fchk
    txt_fchk = open('qchem_temp_32947.fchk', 'r').read()
    parsed_data = parser_fchk(txt_fchk)
    txt_fchk_new = build_fchk(parsed_data)
    with open('test.fchk', 'w') as f:
        f.write(txt_fchk_new)

    #structure, basis, alpha_coeff, beta_coeff = get_data_from_file_fchk('qchem_temp_32947.fchk')
    #print(basis)


import os
from subprocess import Popen, PIPE
import numpy as np
import hashlib
import pickle
import warnings
from pyqchem.qc_input import QchemInput
from pyqchem.errors import ParserError, OutputError

__calculation_data_filename__ = 'calculation_data.pkl'
try:
    with open(__calculation_data_filename__, 'rb') as input:
        calculation_data = pickle.load(input)
        print('Loaded data from {}'.format(__calculation_data_filename__))
except IOError:
    calculation_data = {}


def redefine_calculation_data_filename(filename):
    global __calculation_data_filename__
    global calculation_data

    __calculation_data_filename__ = filename
    print('Set data file to {}'.format(__calculation_data_filename__))

    try:
        with open(__calculation_data_filename__, 'rb') as input:
            calculation_data = pickle.load(input)
            print('Loaded data from {}'.format(__calculation_data_filename__))
    except IOError:
        calculation_data = {}


# Check if calculation finished ok
def finish_ok(output):
    return output[-1000:].find('Thank you very much for using Q-Chem') != -1


# Layer of compatibility with old version
def create_qchem_input(*args, **kwargs):
    return QchemInput(*args, **kwargs)


def parse_output(get_output_function):
    """
    to be deprecated

    :param get_output_function:
    :return:
    """

    global calculation_data

    def func_wrapper(*args, **kwargs):
        parser = kwargs.pop('parser', None)
        parser_parameters = kwargs.pop('parser_parameters', {})
        store_output = kwargs.pop('store_output', None)

        force_recalculation = kwargs.pop('force_recalculation', False)

        if parser is not None:
            hash_p = (args[0], parser.__name__)
            if hash_p in calculation_data and not force_recalculation:
                print('already calculated. Skip')
                return calculation_data[hash_p]

        output, err = get_output_function(*args, **kwargs)

        if store_output is not None:
            with open('{}'.format(store_output), 'w') as f:
                f.write(output)

        if len(err) > 0:
            print(output[-800:])
            print(err)
            raise Exception('q-chem calculation finished with error')

        if parser is None:
            return output

        parsed_output = parser(output, **parser_parameters)

        calculation_data[hash_p] = parsed_output
        with open(__calculation_data_filename__, 'wb') as output:
            pickle.dump(calculation_data, output, pickle.HIGHEST_PROTOCOL)

        return parsed_output

    return func_wrapper


def local_run(input_file_name, work_dir, fchk_file, use_mpi=False, processors=1):
    """
    Run Q-Chem locally

    :param input_file_name: Q-Chem input file in plain text format
    :param work_dir:  Scratch directory where calculation run
    :param fchk_file: filename of fchk
    :param use_mpi: use mpi instead of openmp
    :return output, err: Q-Chem standard output and standard error
    """

    if not use_mpi:
        os.environ["QCTHREADS"] = "{}".format(processors)
        os.environ["OMP_NUM_THREADS"] = "{}".format(processors)
        os.environ["MKL_NUM_THREADS"] = "1"

    os.environ["GUIFILE"] = fchk_file
    qc_dir = os.environ['QC']
    binary = "{}/exe/qcprog.exe".format(qc_dir)
    # command = binary + ' {} {} '.format(flag, processors) + ' {} '.format(temp_file_name)
    command = binary + ' {} '.format(os.path.join(work_dir, input_file_name)) + ' {} '.format(work_dir)

    qchem_process = Popen(command, stdout=PIPE, stdin=PIPE, stderr=PIPE, shell=True, cwd=work_dir)
    (output, err) = qchem_process.communicate()
    qchem_process.wait()
    output = output.decode()
    err = err.decode()

    return output, err


def remote_run(input_file_name, work_dir, fchk_file, remote_params, use_mpi=False, processors=1):
    """
    Run Q-Chem remotely

    :param input_file: Q-Chem input file in plain text format
    :param work_dir:  Scratch directory where calculation run
    :param fchk_file: filename of fchk
    :param remote_params: connection parameters for paramiko
    :param use_mpi: use mpi instead of openmp
    :return output, err: Q-Chem standard output and standard error
    """
    import paramiko

    # get precommands
    commands = remote_params.pop('precommand', [])
    remote_scratch = remote_params.pop('remote_scratch', None)

    # Setup SSH connection
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(**remote_params)

    ssh.get_transport()
    sftp = ssh.open_sftp()
    print('connected to {}..'.format(remote_params['hostname']))

    # Define temp remote dir
    _, stdout, _ = ssh.exec_command('pwd', get_pty=True)

    if remote_scratch is None:
        remote_scratch = stdout.read().decode().strip('\n').strip('\r')

    remote_dir = '{}/temp_pyqchem_remote/'.format(remote_scratch)

    # Create temp directory in remote machine
    try:
        sftp.mkdir(remote_dir)
    except OSError:
        pass
    sftp.chdir(remote_dir)

    # Copy all files in local workdir to remote machine
    file_list = os.listdir(work_dir)
    for file in file_list:
        sftp.put(os.path.join(work_dir, file), '{}'.format(file))

    flag = '-np' if use_mpi else '-nt'

    # Define commands to run Q-Chem in remote machine
    commands += ['cd  {}'.format(remote_dir),  # go to remote work dir
                 'qchem {} {} {}'.format(flag, processors, input_file_name)]  # run qchem

    # Execute command in remote machine
    stdin, stdout, stderr = ssh.exec_command('bash -l -c "{}"'.format(';'.join(commands)), get_pty=True)

    # Reformat output/error files
    output = ''.join(stdout.readlines())
    error = ''.join(stderr.readlines())

    # get files and remove them from remote server
    for file in sftp.listdir():
        sftp.get(os.path.join(remote_dir, file), os.path.join(work_dir, file))
        sftp.remove(os.path.join(remote_dir, file))
    sftp.rmdir(remote_dir)

    sftp.close()
    ssh.close()

    # Rename fchk file to match expected name
    if input_file_name + '.fchk' in os.listdir(work_dir):
        os.rename(os.path.join(work_dir, input_file_name + '.fchk'), os.path.join(work_dir, fchk_file))

    return output, error


def store_calculation_data(input_qchem, keyword, data, protocol=pickle.HIGHEST_PROTOCOL):

    calculation_data[(hash(input_qchem), keyword)] = data
    with open(__calculation_data_filename__, 'wb') as f:
        pickle.dump(calculation_data, f, protocol)


def retrieve_calculation_data(input_qchem, keyword):
    return calculation_data[(hash(input_qchem), keyword)] if (hash(input_qchem), keyword) in calculation_data else None


def get_output_from_qchem(input_qchem,
                          processors=1,
                          use_mpi=False,
                          scratch=None,
                          read_fchk=False,
                          parser=None,
                          parser_parameters=None,
                          force_recalculation=False,
                          fchk_only=False,
                          store_full_output=False,
                          remote=None,
                          strict_policy=False):
    """
    Runs qchem and returns the output in the following format:
    1) If read_fchk is requested:
        [output, error, parsed_fchk]
    2) If read_fchk is not requested:
        [output, error]

    Note: if parser is set then output contains a dictionary with the parsed info
          else output contains the q-chem output in plain text

    error: contains the standard error data from the calculation (if all OK, then should contain nothing)
    read_fchk: contains a dictionary with the parsed info inside fchk file.

    :param input_qchem:
    :param processors:
    :param use_mpi:
    :param scratch:
    :param read_fchk:
    :param parser:
    :param parser_parameters:
    :param force_recalculation:
    :param fchk_only:
    :param remote:
    :param strict_policy:
    :return output, error[, fchk_dict]:
    """
    from pyqchem.parsers.parser_fchk import parser_fchk

    # check gui > 2 if read_fchk
    if read_fchk:
        if input_qchem.gui is None or input_qchem.gui < 1:
            input_qchem.gui = 2

    if scratch is None:
        scratch = os.environ['QCSCRATCH']

    work_dir = '{}/qchem{}/'.format(scratch, os.getpid())

    try:
        os.mkdir(work_dir)
    except FileExistsError:
        pass

    # check scf_guess if guess
    if input_qchem.mo_coefficients is not None:
        guess = input_qchem.mo_coefficients
        # set guess in place
        mo_coeffa = np.array(guess['alpha'], dtype=np.float)
        l = len(mo_coeffa)
        if 'beta' in guess:
            mo_coeffb = np.array(guess['beta'], dtype=np.float)
        else:
            mo_coeffb = mo_coeffa

        mo_ene = np.zeros(l)

        guess_file = np.vstack([mo_coeffa, mo_ene, mo_coeffb, mo_ene]).flatten()
        with open(work_dir + '53.0', 'w') as f:
            guess_file.tofile(f, sep='')

    input_txt = input_qchem.get_txt()

    # check if parameters is None
    if parser_parameters is None:
        parser_parameters = {}

    # check if full output is stored
    # print('input:', input_qchem)
    output, err = calculation_data[(hash(input_qchem), 'fullout')] if (hash(input_qchem), 'fullout') in calculation_data else [None, None]
    # output, err = retrieve_calculation_data(input_qchem, 'fullout') if retrieve_calculation_data(input_qchem, 'fullout') is not None else [None, None]

    if not force_recalculation and not store_full_output:

        data_fchk = retrieve_calculation_data(input_qchem, 'fchk')

        if parser is not None:

            data = retrieve_calculation_data(hash(input_qchem), parser.__name__)

            if data is not None:
                if read_fchk is False:
                    return data
                elif data_fchk is not None:
                    return data, data_fchk

        else:
            if fchk_only and data_fchk is not None:
                return None, data_fchk

    fchk_filename = 'qchem_temp_{}.fchk'.format(os.getpid())
    temp_filename = 'qchem_temp_{}.inp'.format(os.getpid())

    qchem_input_file = open(os.path.join(work_dir, temp_filename), mode='w')
    qchem_input_file.write(input_txt)
    qchem_input_file.close()

    # Q-Chem calculation
    if output is None or force_recalculation is True:
        if remote is None:
            output, err = local_run(temp_filename, work_dir, fchk_filename, use_mpi=use_mpi, processors=processors)
        else:
            output, err = remote_run(temp_filename, work_dir, fchk_filename, remote, use_mpi=use_mpi, processors=processors)

    if not finish_ok(output):
        err += '\n'.join(output.split('\n')[-10:])
        raise OutputError(output, err)

    if store_full_output:
        store_calculation_data(input_qchem, 'fullout', [output, err])

    if parser is not None:
        try:
            output = parser(output, **parser_parameters)
        # minimum functionality for error capture
        except:
            raise ParserError(parser.__name__, 'general')

        store_calculation_data(input_qchem, parser.__name__, output)

    if read_fchk:

        data_fchk = retrieve_calculation_data(input_qchem, 'fchk')
        if data_fchk is not None and not force_recalculation:
            return output, data_fchk

        if not os.path.isfile(os.path.join(work_dir, fchk_filename)):
            warnings.warn('fchk not found! Make sure the input generates it (gui 2)')
            return output, []

        with open(os.path.join(work_dir, fchk_filename)) as f:
            fchk_txt = f.read()
        os.remove(os.path.join(work_dir, fchk_filename))

        data_fchk = parser_fchk(fchk_txt)
        store_calculation_data(input_qchem, 'fchk', data_fchk)

        return output, data_fchk

    return output


def get_input_hash(data):
    return hashlib.md5(data.encode()).hexdigest()




import requests as req
from lxml import html
import unicodedata
import re
import numpy as np
from copy import deepcopy


def _txt_to_basis_dict(basis_txt):
    # read basis in gaussian/qchem format

    symbol = basis_txt[0].split()[0]

    def is_number(s):
        try:
            float(s)
            return True
        except ValueError:
            return False

    basis_pure = basis_txt[2:]

    section_marks = []
    for i, line in enumerate(basis_pure):
        if not is_number(line.split()[0]):
            section_marks.append(i)

    shells = []
    for i in section_marks[:-1]:
        type, n_func, _  = basis_pure[i].split()
        n_func = int(n_func)

        if type.upper() in ['SP']:
            p_exponent, con_coefficients, p_con_coefficients = np.array([line.split()
                                                                         for line in basis_pure[i + 1:i + n_func + 1]],
                                                                        dtype=float).T
        else:
            p_exponent, con_coefficients = np.array([line.split()
                                                     for line in basis_pure[i + 1:i + n_func + 1]],
                                                    dtype=float).T
            p_con_coefficients = np.zeros_like(p_exponent)


        shells.append({'shell_type': type,
                       'p_exponents': list(p_exponent),
                       'con_coefficients': list(con_coefficients),
                       'p_con_coefficients': list(p_con_coefficients)})

    return {'symbol': symbol,
            'shells': shells}


def get_basis_element_from_ccRepo(element,
                                  program='Gaussian',
                                  basis='cc-pVDZ'):

    # Check main page element list
    with req.get("http://www.grant-hill.group.shef.ac.uk/ccrepo/") as resp:
        element_list = []
        for m in re.finditer('class="sym" href=', resp.text):
            # print(resp.text[m.end():m.end()+50].split('"')[1] )
            element_list.append(resp.text[m.end():m.end()+50].split('"')[1])

    # define symbol to element dictionary (more should be added)
    element_dict = {'H': 'hydrogen',
                    'He': 'helum',
                    'Li': 'lithium',
                    'Be': 'beryllium',
                    'B': 'boron',
                    'C': 'carbon',
                    'N': 'nitrogen',
                    'O': 'oxygen',
                    'F': 'fluorine',
                    'S': 'sulfur',
                    'Cl': 'chlorine',
                    'Si': 'silicon',
                    'Se': 'selenium',
                    'Br': 'bromine',
                    'I': 'iodine'
                    }

    resp = req.get("http://www.grant-hill.group.shef.ac.uk/ccrepo/{}".format(element_dict[element]))
    n_ini = resp.text.find('form-inline')

    # determinte the php web address for particular element
    php_site = resp.text[n_ini:n_ini+100].split('"')[2]

    # print('http://www.grant-hill.group.shef.ac.uk/ccrepo/{}/{}'.format(element, php_site))
    # r = req.post(url = 'http://www.grant-hill.group.shef.ac.uk/ccrepo/titanium/tibasis.php', data = data)
    r = req.post(url='http://www.grant-hill.group.shef.ac.uk/ccrepo/{}/{}'.format(element_dict[element], php_site),
                 data={'basis': basis, 'program': program})

    # parse the data
    tree = html.fromstring(r.content)
    r.close()

    header = tree.xpath('//div[@class="container"]/text()')
    citation = unicodedata.normalize('NFC', header[1]).strip()
    description = unicodedata.normalize('NFC', header[2]).strip()

    basis_data = tree.xpath('/html/body/div/nobr/text()')
    basis_clean = [unicodedata.normalize('NFKC', line).strip() for line in basis_data]
    # basis_name = basis_clean[1].split('"')[1]

    return citation, description, basis_clean[2:]


def get_basis_from_ccRepo(structure, basis, full=False, if_missing=None):

    symbols = structure.get_atomic_elements()
    if not full:
        symbols = np.unique(symbols)

    atoms = []
    for symbol in symbols:
        citation, description, basis_data = get_basis_element_from_ccRepo(symbol,
                                                                          program='Gaussian',
                                                                          basis=basis)

        if len(basis_data) == 0:
            if if_missing is not None:
                citation, description, basis_data = get_basis_element_from_ccRepo(symbol,
                                                                                  program='Gaussian',
                                                                                  basis=if_missing)
                if len(basis_data) == 0:
                    raise Exception('Basis {}, {} not found for atom {}'.format(basis, if_missing, symbol))
            else:
                raise Exception('Basis {} not found for atom {}'.format(basis, symbol))

        # print(description)
        # print(citation)
        atoms.append(_txt_to_basis_dict(basis_data))

    basis_set = {'name': basis,
                 'primitive_type': 'gaussian',
                 'atoms': atoms}

    return basis_set


def basis_to_txt(basis):
    # write basis in qchem/gaussian format
    basis_txt = ''

    for atom in basis['atoms']:
        basis_txt += atom['symbol'] + '\n'
        for shell in atom['shells']:
            basis_txt += '{} {} {}\n'.format(shell['shell_type'].upper(), len(shell['p_exponents']), 1.00)
            for p, c, pc in zip(shell['p_exponents'], shell['con_coefficients'], shell['p_con_coefficients']):
                if shell['shell_type'].upper() in ['SP']:
                    basis_txt += '{:15.10e} {:15.10e} {:15.10e} \n'.format(p, c, pc)
                else:
                    basis_txt += '{:15.10e} {:15.10e} \n'.format(p, c)

        basis_txt += '****\n'
    return basis_txt


def trucate_basis(basis, shells=()):

    basis_trunc = deepcopy(basis)
    for check_shell in shells:
        for atom in basis_trunc['atoms']:
            delete_list = []
            for i, shell in enumerate(atom['shells']):
                if shell['shell_type'].upper() == check_shell.upper():
                    del shell
                    delete_list.append(i)
            for i in sorted(delete_list, reverse=True):
                del atom['shells'][i]

    return basis_trunc

if __name__ == '__main__':

    citation, description, basis = get_basis_element_from_ccRepo('C',
                                                                 program='Gaussian',
                                                                 basis='cc-pVTZ')

    print(citation)
    print(description)
    print('-----------------------')
    for line in basis:
        print(line)

    basis = _txt_to_basis_dict(basis)
    print(basis)


    from pyqchem.structure import Structure


    # create molecule
    molecule = Structure(coordinates=[[0.0, 0.0, 0.0000],
                                      [0.0, 0.0, 1.5811]],
                         atomic_elements=['Se', 'H'],
                         charge=-1,
                         multiplicity=1)

    basis = get_basis_from_ccRepo(molecule,
                                  basis='cc-pVTZ',
                                  full=False)

    print(basis_to_txt(basis))



import numpy as np
import matplotlib.pyplot as plt


def draw_electron(pos, spin=1, color='black'):
    a = 0.5
    s = 0.05
    if spin == 1:
        d = 0.1
        plt.plot([pos[0]+d, pos[0]+0.1+d], [pos[1]+a, pos[1]+a*0.6], 'k-', lw=2, color=color)
    else:
        d = -0.1
        plt.plot([pos[0]+d, pos[0]-0.1+d], [pos[1]+s, pos[1]+a*0.4+s], 'k-', lw=2, color=color)

    plt.plot([pos[0]+d, pos[0]+d], [pos[1]+s, pos[1] + a], 'k-', lw=2, color=color)


def plot_state(alpha, beta, index=1):

    levels_alpha = range(len(alpha))

    plt.grid(axis='y')

    x = [index + 1] * len(levels_alpha)

    plt.annotate(index+1, xy=(index+1, 0), xytext=(0, -4), size=10,
                 ha="center", va="top", textcoords="offset points")

    for xi, yi, oa, ob in zip(x,levels_alpha, alpha, beta):

        plt.plot([xi-0.3, xi+0.3], [yi, yi], 'k-', lw=3, color='blue')
        if oa == '1':
            draw_electron([xi, yi], spin=1, color='r')
        if ob == '1':
            draw_electron([xi, yi], spin=-1, color='r')

    plt.margins(0.2)





import re
import numpy as np


# parser for frequencies calculations
def basic_frequencies(output, print_data=False):
    """
    Parser for frequencies calculations

    :param output:
    :param print_data:
    :return:
    """

    # Coordinates
    n = output.find('$molecule')
    n2 = output[n:].find('$end')
    molecule_region = output[n:n+n2-1].replace('\t', ' ').split('\n')[1:]
    coordinates = np.array([ np.array(line.split()[1:4], dtype=float) for line in molecule_region[1:]])
    symbols = [line.split()[0].capitalize() for line in molecule_region[1:]]
    n_atoms = len(coordinates)

    # Energy
    n = output.find('Total energy in the final basis set =')
    energy = float(output[n:n+70].split()[8])

    n = output.find('VIBRATIONAL ANALYSIS')
    vibration_section = output[n:]

    modes = []
    frequencies = []
    force_constants = []
    red_mass = []
    ir_active = []
    ir_intens = []
    raman_active = []
    for m in re.finditer('Mode:', vibration_section):
        end_line = vibration_section[m.end():].find('\n')
        modes += vibration_section[m.end():m.end()+end_line].split()[:3]

    for m in re.finditer('Frequency:', vibration_section):
        end_line = vibration_section[m.end():].find('\n')
        frequencies += vibration_section[m.end():m.end()+end_line].split()[:3]

    for m in re.finditer('Force Cnst:', vibration_section):
        end_line = vibration_section[m.end():].find('\n')
        force_constants += vibration_section[m.end():m.end()+end_line].split()[:3]

    for m in re.finditer('Red. Mass:', vibration_section):
        end_line = vibration_section[m.end():].find('\n')
        red_mass += vibration_section[m.end():m.end()+end_line].split()[:3]

    for m in re.finditer('IR Active:', vibration_section):
        end_line = vibration_section[m.end():].find('\n')
        ir_active += vibration_section[m.end():m.end()+end_line].split()[:3]

    for m in re.finditer('IR Intens:', vibration_section):
        end_line = vibration_section[m.end():].find('\n')
        ir_intens += vibration_section[m.end():m.end()+end_line].split()[:3]

    for m in re.finditer('Raman Active:', vibration_section):
        end_line = vibration_section[m.end():].find('\n')
        raman_active += vibration_section[m.end():m.end()+end_line].split()[:3]

    modes = [int(n) for n in modes]
    frequencies = [float(n) for n in frequencies]
    force_constants = [float(n) for n in force_constants]
    red_mass = [float(n) for n in red_mass]
    ir_active = [bool(n) for n in ir_active]
    ir_intens = [float(n) for n in ir_intens]
    raman_active = [bool(n) for n in raman_active]

    nm_coordinates = []
    for i, line in enumerate(vibration_section.split('\n')):
        if 'X      Y      Z' in line:
            nm_coordinate = []
            for j in range(n_atoms):
                coor_lines = vibration_section.split('\n')[j+ i+ 1]
                nm_coordinate.append(coor_lines.split()[1:])

            nm_coordinate = np.array(nm_coordinate, dtype=float)#.reshape(n_atoms, -1)
            # print(nm_coordinate.shape[1], nm_coordinate.shape[1]//3)

            nm_coordinates += [nm_coordinate[:, i*3:(i+1)*3].tolist() for i in range(nm_coordinate.shape[1]//3)]

    return {'modes': modes,
            'frequencies': frequencies,
            'force_constants': force_constants,
            'red_mass': red_mass,
            'ir_active': ir_active,
            'ir_intens': ir_intens,
            'raman_active': raman_active,
            'coordinates': nm_coordinates,
            'energy': energy}



import re
import numpy as np
from scipy.optimize import leastsq


def standardize_vector(vector):
    import numpy as np
    if vector[0] != 0:
        if vector[0] < 0:
            vector = np.array(vector) * -1
            vector = vector.tolist()
    elif vector[1] != 0:
        if vector[1] < 0:
            vector = np.array(vector) * -1
            vector = vector.tolist()
    else:
        if vector[2] < 0:
            vector = np.array(vector) * -1
            vector = vector.tolist()

    for i in range(3):
        vector[i] = vector[i] + 0

    return vector


def search_bars(output, from_position=0, bar_type='---'):
    output = output[from_position:]
    positions = []
    previous = 0
    for m in re.finditer(bar_type, output):
        if m.start() > previous + 1:
            positions.append(m.start() + from_position)
        previous = m.end()

    return positions


def is_transition(configuration, reference, n_electron=1, max_jump=10):
    """
    Determine if a configuration correspond to a transition of n_electron
    :param configuration: dictionary containing the configuration to be analyzed
    :param reference: reference configuration (in general lowest energy Slater determinant)
    :param n_electron:
    :param max_jump: Restrict to transitions with jumps less or equal to max_jump orbitals
    :return: True if conditions are met, otherwise False
    """

    alpha_diff = [int(i) - int(j) for i, j in zip(configuration['alpha'], reference['alpha'])]
    beta_diff = [int(i) - int(j) for i, j in zip(configuration['beta'], reference['beta'])]

    ini_alpha = np.where(np.array(alpha_diff) < 0)[0]
    fin_alpha = np.where(np.array(alpha_diff) > 0)[0]

    ini_beta = np.where(np.array(beta_diff) < 0)[0]
    fin_beta = np.where(np.array(beta_diff) > 0)[0]

    try:
        jump_alpha = np.max(fin_alpha) - np.min(ini_alpha)
    except ValueError:
        jump_alpha = 0

    try:
        jump_beta = np.max(fin_beta) - np.min(ini_beta)
    except ValueError:
        jump_beta = 0

    n_alpha = len(fin_alpha)
    n_beta = len(fin_beta)

    elec_condition = n_alpha + n_beta == n_electron
    jump_condition = jump_alpha <= max_jump and jump_beta <= max_jump

    return elec_condition and jump_condition


def get_ratio_of_condition(state, n_electron=1, max_jump=10):
    # reference = {'hole': '', 'alpha': '111000', 'beta': '111000', 'part': '', 'amplitude': 0.9777044}

    alpha_int = [int(i) for i in state['configurations'][0]['alpha']]
    ref_alpha = ''
    for i in range(len(alpha_int)):
        if i < np.sum(alpha_int):
            ref_alpha += '1'
        else:
            ref_alpha += '0'

    reference = {'hole': '', 'alpha': ref_alpha, 'beta': ref_alpha, 'part': ''}

    p = 0
    for configuration in state['configurations']:
        if is_transition(configuration, reference, n_electron, max_jump):
            p += configuration['amplitude']**2

    return p


def set_zero_coefficients(basis, mo_coeff, range_atoms):
    """
    set 0.0 to coefficients of functions centered in atoms 'range_atoms'

    :param basis: full basis dictionary
    :param mo_coeff: full molecular orbitals coefficients to be modiffied
    :param range_atoms: list containing the atom numbers whose coefficients will be set to zero
    :return:
    """

    functions_to_atom = []
    nat = len(basis['atoms'])
    for i in range(0, nat):
        nf = 0
        for shell in basis['atoms'][i]['shells']:
            nf += shell['functions']
        functions_to_atom.append(nf)
    functions_to_atom = functions_to_atom

    # print(funtions_to_atom)
    mo_coeff_a = np.array(mo_coeff['alpha'])
    for i in range_atoms:
        ini = np.sum(functions_to_atom[:i], dtype=int)
        fin = np.sum(functions_to_atom[:i+1], dtype=int)
        # print('ini', ini, 'fin', fin)
        mo_coeff_a[:, ini: fin] *= 0.0
    mo_coeff_zero = {'alpha': mo_coeff_a.tolist()}

    if 'beta' in mo_coeff:
        mo_coeff_b = np.array(mo_coeff['beta'])
        for i in range_atoms:
            ini = np.sum(functions_to_atom[:i], dtype=int)
            fin = np.sum(functions_to_atom[:i + 1], dtype=int)
            # print('ini', ini, 'fin', fin)
            mo_coeff_b[:, ini: fin] *= 0.0
        mo_coeff_zero['beta'] = mo_coeff_b.tolist()

    return mo_coeff_zero


def get_plane(coords, direction=None):
    """Generate initial guess"""

    coords = np.array(coords)
    p0 = np.cross(coords[0] - coords[2], coords[0] - coords[-1]) / np.linalg.norm(np.cross(coords[0] - coords[2],
                                                                                           coords[0] - coords[-1]))

    # Fitting function to a plane
    def fitfunc(p, coords):
        average = np.average(coords, axis=0)
        return np.array([np.dot(p, average - c) for c in coords])

    # Error function (including force norm(normal) = 1)
    errfunc = lambda p, x: fitfunc(p, x)**2 + (np.linalg.norm(p) - 1.0)**2

    p1, flag = leastsq(errfunc, p0, args=(coords,))


    # Check final result
    point = np.average(coords, axis=0).tolist()
    normal = np.array(p1)/np.linalg.norm(p1)

    if direction is not None:
        vector = coords[direction[1]] - coords[direction[0]]
        # proj = vector - np.dot(vector, normal)*normal
        projected = np.cross(normal,  np.cross(vector, normal))
        projected /= np.linalg.norm(projected)

        normal = standardize_vector(normal)
        projected = standardize_vector(projected)

        return point, normal, projected

    normal = standardize_vector(normal)
    return point, normal


def reorder_coefficients(occupations, coefficients):
    """
    Reorder the coefficients according to occupations. Occupated orbitals will be grouped at the beginning
    non occupied will be attached at the end

    :param occupations: list on integers (0 or 1) or list of Boolean
    :param coefficients:  dictionary containing the molecular orbitals coefficients {'alpha': coeff, 'beta:' coeff}.
                          coeff should be a list of lists (Norb x NBas)
    :return:
    """

    alpha_coefficients = []
    non_occupied = []
    for occ, coeff in zip(occupations['alpha'], coefficients['alpha']):
        if occ:
            alpha_coefficients.append(coeff)
        else:
            non_occupied.append(coeff)

    # non occupated attached at the end
    alpha_coefficients += non_occupied

    if not 'beta' in coefficients:
        coefficients['beta'] = coefficients['alpha']

    beta_coefficients = []
    non_occupied = []
    for occ, coeff in zip(occupations['beta'], coefficients['beta']):
        if occ:
            beta_coefficients.append(coeff)
        else:
            non_occupied.append(coeff)

    # non occupated attached at the end
    beta_coefficients += non_occupied

    return {'alpha': alpha_coefficients, 'beta': beta_coefficients}


def get_occupied_electrons(configuration, structure):
    # not working for particle, hole enabled configurations!
    alpha_e = np.sum([int(c) for c in configuration['alpha']])
    beta_e = np.sum([int(c) for c in configuration['beta']])
    return (structure.number_of_electrons + structure.charge - (alpha_e + beta_e))//2


if __name__ == '__main__':
    state = {'configurations': [{'hole': '', 'alpha': '110100', 'beta': '111000', 'part': '', 'amplitude': 0.5}]}
    b = get_ratio_of_condition(state)
    print('test', b)


__author__ = 'Abel Carreras'

import re
import operator
from pyqchem.utils import standardize_vector
from pyqchem.structure import Structure



def basic_rasci(output):
    """
    Basic Parser for RAS-CI calculations

    :param output:
    :return:
    """

    # scf energy
    enum = output.find('SCF   energy in the final basis set')
    scf_energy = float(output[enum:enum+100].split()[8])

    # excited states data
    excited_states = []
    for m in re.finditer('RAS-CI total energy for state', output):
        # print('ll found', m.start(), m.end())

        section_state = output[m.end():m.end() + 10000]  # 10000: assumed to max of section
        section_state = section_state[:section_state.find('********')]

        enum = section_state.find('RAS-CI total energy for state')
        section_state = section_state[:enum]

        # energies
        tot_energy = float(section_state.split()[1])
        exc_energy_units = section_state.split()[4][1:-1]
        exc_energy = float(section_state.split()[6])
        mul = section_state.split()[8]

        # dipole moment
        enum = section_state.find('Dipole Moment')
        dipole_mom = [float(section_state[enum:].split()[2]) + 0.0,
                      float(section_state[enum:].split()[4]) + 0.0,
                      float(section_state[enum:].split()[6]) + 0.0]

        # Transition moment
        enum = section_state.find('Trans. Moment')
        if enum > -1:
            trans_mom = [float(section_state[enum:].split()[2]) + 0.0,
                         float(section_state[enum:].split()[4]) + 0.0,
                         float(section_state[enum:].split()[6]) + 0.0]
            trans_mom = standardize_vector(trans_mom)
            strength = float(section_state[enum:].split()[10])
        else:
            trans_mom = None
            strength = None

        # amplitudes table
        enum = section_state.find('AMPLITUDE')
        enum2 = section_state.find('Contributions')
        section_table = section_state[enum: enum2].split('\n')[2:-2]

        # ' HOLE  | ALPHA | BETA  | PART | AMPLITUDE'

        table = []
        for row in section_table:
            table.append({'hole': row.split('|')[1].strip(),
                          'alpha': row.split('|')[2].strip(),
                          'beta': row.split('|')[3].strip(),
                          'part': row.split('|')[4].strip(),
                          'amplitude': float(row.split('|')[5]) + 0.0})
        table = sorted(table, key=operator.itemgetter('hole', 'alpha', 'beta', 'part'))

        # Contributions RASCI wfn
        contributions_section = section_state[enum2:]
        contributions = {'active' : float(contributions_section.split()[4]),
                         'hole': float(contributions_section.split()[6]),
                         'part': float(contributions_section.split()[8])}

        # complete dictionary
        tot_energy_units = 'au'
        excited_states.append({'total_energy': tot_energy,
                               'total energy units': tot_energy_units,
                               'excitation_energy': exc_energy,
                               'excitation energy units': exc_energy_units,
                               'multiplicity': mul,
                               'dipole_moment': dipole_mom,
                               'transition_moment': trans_mom,
                               # 'oscillator_strength': strength,
                               'configurations': table,
                               'contributions_fwn': contributions})

    return {'scf energy': scf_energy,
            'excited states rasci': excited_states}



__author__ = 'Abel Carreras'

import re
import operator
from pyqchem.utils import standardize_vector
from pyqchem.structure import Structure
from pyqchem.utils import search_bars

def _read_simple_matrix(header, output, maxchar=10000, foot='-------'):
    matrix_list = []
    for m in re.finditer(header, output):
        section_state = output[m.end():m.end() + maxchar]  # 10000: assumed to max of section
        section_state = section_state[:section_state.find(foot)]
        dim = len(section_state.split('\n')[1].split())
        matrix = section_state.split('\n')[1:dim + 1]
        matrix = [[float(n) for n in l.split()] for l in matrix]
        matrix_list.append(matrix)

    return matrix_list


def _read_soc_matrix(lines, dimensions):
    matrix = []
    for ib in range(dimensions[0]):
        real = []
        complex = []
        for j in range((dimensions[1] + 1) // 2):
            real += lines[j*3 + 2 + ib].split()[1:][0::2]
            complex += lines[j*3 +2+ib].split()[1:][1::2]

        row = [float(r) + float(c) * 1j for r, c in zip(real, complex)]
        matrix.append(row)

    return matrix


def rasci(output):
    """
    Parser for RAS-CI calculations
    Include:
    - Diabatization scheme data
    - Structure
    - Adiabatic states

    :param output:
    :return:
    """

    data_dict = {}
    # Molecule
    n = output.find('$molecule')
    n2 = output[n:].find('$end')

    molecule_region = output[n:n+n2-1].replace('\t', ' ').split('\n')[1:]
    charge, multiplicity = [int(num) for num in molecule_region[0].split()]
    coordinates = [[float(l) for l in line.split()[1:4]] for line in molecule_region[1:]]
    symbols = [line.split()[0].capitalize() for line in molecule_region[1:]]
    n_atoms = len(symbols)

    # structure
    structure_input = Structure(coordinates=coordinates,
                                atomic_elements=symbols,
                                charge=charge,
                                multiplicity=multiplicity)

    enum = output.find('Standard Nuclear Orientation')
    section_structure = output[enum:enum + 100*structure_input.get_number_of_atoms()].split('\n')
    section_structure = section_structure[3:structure_input.get_number_of_atoms()+3]
    coordinates = [ [float(num) for num in s.split()[2:]] for s in section_structure]

    data_dict['structure'] = Structure(coordinates=coordinates,
                                       atomic_elements=symbols,
                                       charge=charge,
                                       multiplicity=multiplicity)

    # scf energy
    enum = output.find('SCF   energy in the final basis set')
    scf_energy = float(output[enum:enum+100].split()[8])

    data_dict['scf energy'] = scf_energy
    # total energy
    # enum = output.find('Total energy in the final basis set')
    # total_energy = float(output[enum:enum+100].split()[8])

    # Diabatization scheme
    done_diabat = bool(output.find('RASCI DIABATIZATION')+1)
    if done_diabat:
        rot_matrix = _read_simple_matrix('showmatrix final adiabatic -> diabatic', output)[-1]
        adiabatic_matrix = _read_simple_matrix('showing H in adiabatic representation: NO coupling elements', output)[-1]
        diabatic_matrix = _read_simple_matrix('showing H in diabatic representation: WITH coupling elements', output)[-1]

        mulliken_adiabatic = []
        enum = output.find('Mulliken analysis of Adiabatic State')
        for m in re.finditer('Mulliken analysis of Adiabatic State', output[enum:]):
            section_mulliken = output[m.end() + enum: m.end() + 10000 + enum]  # 10000: assumed to max of section
            section_mulliken = section_mulliken[:section_mulliken.find('Natural Orbitals stored in FCHK')]
            section_attachment = section_mulliken.split('\n')[9+n_atoms:9+n_atoms*2]

            mulliken_adiabatic.append({'attach': [float(l.split()[1]) for l in section_attachment],
                                       'detach': [float(l.split()[2]) for l in section_attachment],
                                       'total': [float(l.split()[3]) for l in section_attachment]})

        mulliken_diabatic = []
        enum = output.find('showing H in diabatic representation')
        for m in re.finditer('Mulliken Analysis of Diabatic State', output[enum:]):
            section_mulliken = output[m.end() + enum: m.end() + 10000 + enum]  # 10000: assumed to max of section
            section_mulliken = section_mulliken[:section_mulliken.find('Natural Orbitals stored in FCHK')]
            section_attachment = section_mulliken.split('\n')[9+n_atoms:9+n_atoms*2]

            mulliken_diabatic.append({'attach': [float(l.split()[1]) for l in section_attachment],
                             'detach': [float(l.split()[2]) for l in section_attachment],
                             'total': [float(l.split()[3]) for l in section_attachment]})

        data_dict['diabatization'] = {'rot_matrix': rot_matrix,
                                      'adiabatic_matrix': adiabatic_matrix,
                                      'diabatic_matrix': diabatic_matrix,
                                      'mulliken_analysis': mulliken_diabatic,
                                      'mulliken_adiabatic': mulliken_adiabatic}

    # excited states data
    excited_states = []
    for m in re.finditer('RAS-CI total energy for state', output):
        # print('ll found', m.start(), m.end())

        section_state = output[m.end():m.end() + 10000]  # 10000: assumed to max of section
        section_state = section_state[:section_state.find('********')]

        enum = section_state.find('RAS-CI total energy for state')
        section_state = section_state[:enum]

        # energies
        tot_energy = float(section_state.split()[1])
        exc_energy_units = section_state.split()[4][1:-1]
        exc_energy = float(section_state.split()[6])
        mul = section_state.split()[8]

        # dipole moment
        enum = section_state.find('Dipole Moment')
        dipole_mom = [float(section_state[enum:].split()[2]) + 0.0,
                      float(section_state[enum:].split()[4]) + 0.0,
                      float(section_state[enum:].split()[6]) + 0.0]

        # Transition moment
        enum = section_state.find('Trans. Moment')
        if enum > -1:
            trans_mom = [float(section_state[enum:].split()[2]) + 0.0,
                         float(section_state[enum:].split()[4]) + 0.0,
                         float(section_state[enum:].split()[6]) + 0.0]
            trans_mom = standardize_vector(trans_mom)
            strength = float(section_state[enum:].split()[10])
        else:
            trans_mom = None
            strength = None

        # configurations table
        enum = section_state.find('AMPLITUDE')
        enum2 = section_state.find('Contributions')
        section_table = section_state[enum: enum2].split('\n')[2:-2]

        # ' HOLE  | ALPHA | BETA  | PART | AMPLITUDE'

        table = []
        for row in section_table:
            table.append({'hole': row.split('|')[1].strip(),
                          'alpha': row.split('|')[2].strip(),
                          'beta': row.split('|')[3].strip(),
                          'part': row.split('|')[4].strip(),
                          'amplitude': float(row.split('|')[5]) + 0.0})
        table = sorted(table, key=operator.itemgetter('hole', 'alpha', 'beta', 'part'))

        # Contributions RASCI wfn
        contributions_section = section_state[enum2:]
        contributions = {'active' : float(contributions_section.split()[4]),
                         'hole': float(contributions_section.split()[6]),
                         'part': float(contributions_section.split()[8])}

        # complete dictionary
        tot_energy_units = 'au'
        excited_states.append({'total_energy': tot_energy,
                               'total energy units': tot_energy_units,
                               'excitation_energy': exc_energy,
                               'excitation energy units': exc_energy_units,
                               'multiplicity': mul,
                               'dipole_moment': dipole_mom,
                               'transition_moment': trans_mom,
                               'oscillator_strength': strength,
                               'configurations': table,
                               'contributions_fwn': contributions})

    data_dict.update({'excited states rasci': excited_states})

    # Interstate transition properties
    done_interstate = bool(output.find('Interstate Transition Properties')+1)
    if done_interstate:
        ini_section = output.find('Interstate Transition Properties')
        end_section = search_bars(output, from_position=ini_section)[1]
        interstate_section = output[ini_section: end_section]

        interstate_dict = {}
        for m in re.finditer('State A: Root', interstate_section):
            section_pair = interstate_section[m.start():m.start() + 10000]
            section_pair = section_pair[:section_pair.find('********')]
            # print(section_pair)
            lines = section_pair.split('\n')

            state_a = int(lines[0].split()[-1])
            state_b = int(lines[1].split()[-1])

            pair_dict = {'state_a': state_a,
                         'state_b': state_b}

            s_a = s_b = 0
            for i, line in enumerate(lines):
                # RAS-CI SOC section
                if '||gamma^AB||_total' in line:
                    pair_dict['gamma_total'] = float(lines[i+0].split()[-1])
                    pair_dict['gamma_sym'] = float(lines[i+1].split()[-1])
                    pair_dict['gamma_anti_sym'] = float(lines[i+2].split()[-1])

                if 'KET: S,Sz' in line:
                    s_a = float(lines[i].split()[-2])
                    s_b = float(lines[i+1].split()[-2])
                if '1-elec SOC matrix (cm-1)' in line:
                    pair_dict['1e_soc_mat'] = _read_soc_matrix(lines[i+1:], [int(2*s_b + 1), int(2*s_a+1)])
                if '2e-SOMF Reduced matrix elements (cm-1)' in line:
                    r, c = lines[i+1].split()[-2:]
                    pair_dict['hso_l-'] = float(r) + float(c) * 1j
                    r, c = lines[i+2].split()[-2:]
                    pair_dict['hso_l0'] = float(r) + float(c) * 1j
                    r, c = lines[i+3].split()[-2:]
                    pair_dict['hso_l+'] = float(r) + float(c) * 1j

                if '2-elec mean-field SOC matrix (cm-1)' in line:
                    pair_dict['2e_soc_mat'] = _read_soc_matrix(lines[i + 1:], [int(2 * s_b + 1), int(2 * s_a + 1)])
                if 'Total mean-field SOC matrix (cm-1' in line:
                    pair_dict['total_soc_mat'] = _read_soc_matrix(lines[i + 1:], [int(2 * s_b + 1), int(2 * s_a + 1)])
                if 'Mean-Field SOCC' in line:
                    pair_dict['mf_socc'] = float(line.split()[-2])
                    pair_dict['units'] = line.split()[-1]

            interstate_dict[(state_a, state_b)] = pair_dict
        data_dict.update({'interstate_properties': interstate_dict})

    return data_dict



import numpy as np

def trunc_dictionary_list(dic_data, decimal):
    # w : dictionary
    # decimal: number of decimal places to leave after truncation
    def round_float(num):
        if isinstance(num, float):
            return int(num * 10**decimal)
        else:
            return num

    def iterlist(d):
        lf = []
        for i, v in enumerate(d):
            if isinstance(v, dict):
                iterdict(v)
            elif isinstance(v, list):
                iterlist(v)
            else:
                d[i] = round_float(v)

    def iterdict(d):
        lf = []
        for k,v in d.items():
            if isinstance(v, dict):
                iterdict(v)
            elif isinstance(v, list):
                iterlist(v)
            else:
                d[k] = round_float(v)

    if isinstance(dic_data, dict):
        iterdict(dic_data)
    elif isinstance(dic_data, list):
        iterlist(dic_data)
    else:
        round_float(dic_data)


def standardize_dictionary(dic_data, decimal=5):

    # set all amplitudes to absolute value
    for state in dic_data['excited states rasci']:
        for configuration in state['configurations']:
            configuration['amplitude'] = abs(configuration['amplitude'])

    trunc_dictionary_list(dic_data, decimal)

    # delete zero amplitude configurations
    for state in dic_data['excited states rasci']:
        delete_list = []
        for i, configuration in enumerate(state['configurations']):
            if configuration['amplitude'] == 0:
                delete_list.append(i)
        for i in reversed(delete_list):
            del state['configurations'][i]

    return dic_data


def assertDeepAlmostEqual(test_case, expected, actual, *args, **kwargs):
    """
    Assert that two complex structures have almost equal contents.

    Compares lists, dicts and tuples recursively. Checks numeric values
    using test_case's :py:meth:`unittest.TestCase.assertAlmostEqual` and
    checks all other values with :py:meth:`unittest.TestCase.assertEqual`.
    Accepts additional positional and keyword arguments and pass those
    intact to assertAlmostEqual() (that's how you specify comparison
    precision).

    :param test_case: TestCase object on which we can call all of the basic
    'assert' methods.
    :type test_case: :py:class:`unittest.TestCase` object
    """

    import numpy
    from numpy import long
    is_root = not '__trace' in kwargs
    trace = kwargs.pop('__trace', 'ROOT')
    try:
        if isinstance(expected, (int, float, long, complex)):
            test_case.assertAlmostEqual(expected, actual, *args, **kwargs)
        elif isinstance(expected, (list, tuple, numpy.ndarray)):
            test_case.assertEqual(len(expected), len(actual))
            for index in range(len(expected)):
                v1, v2 = expected[index], actual[index]
                assertDeepAlmostEqual(test_case, v1, v2,
                                      __trace=repr(index), *args, **kwargs)
        elif isinstance(expected, dict):
            test_case.assertEqual(set(expected), set(actual))
            for key in expected:
                assertDeepAlmostEqual(test_case, expected[key], actual[key],
                                      __trace=repr(key), *args, **kwargs)
        else:
            test_case.assertEqual(expected, actual)
    except AssertionError as exc:
        exc.__dict__.setdefault('traces', []).append(trace)
        if is_root:
            trace = ' -> '.join(reversed(exc.traces))
            exc = AssertionError("%s\nTRACE: %s" % (exc.message, trace))
        raise exc



import numpy as np
from copy import deepcopy
from pyqchem.basis import basis_to_txt
import hashlib, json
import warnings
from pyqchem.errors import QchemInputWarning


class QchemInput:
    """
    Handles the Q-Chem input info
    """

    def __init__(self, molecule,
                 jobtype='sp',
                 method='HF',
                 exchange=None,
                 correlation=None,
                 unrestricted=None,
                 basis='6-31G',
                 thresh=14,
                 scf_convergence=8,
                 max_scf_cycles=50,
                 purecart=None,
                 # RASCI
                 ras_roots=1,
                 ras_do_hole=True,
                 ras_do_part=True,
                 ras_act=None,
                 ras_elec=None,
                 ras_elec_alpha=None,
                 ras_elec_beta=None,
                 ras_occ=None,
                 ras_spin_mult=1,
                 ras_sts_tm=False,
                 ras_natorb=False,
                 ras_natorb_state=None,
                 ras_print=1,
                 ras_diabatization_scheme=None,
                 ras_diabatization_states=None,
                 # RASCI SrDFT
                 ras_omega=400,
                 ras_srdft=None,
                 ras_srdft_damp=0.5,
                 ras_srdft_exc=None,
                 ras_srdft_cor=None,
                 ras_srdft_spinpol=0,
                 # SOC
                 calc_soc=False,
                 # cis
                 cis_convergence=6,
                 cis_n_roots=None,
                 cis_singlets=False,
                 cis_triplets=False,
                 cis_ampl_anal=False,
                 loc_cis_ov_separate=False,
                 er_cis_numstate=0,
                 cis_diabath_decompose=False,
                 max_cis_cycles=30,
                 localized_diabatization=None,
                 sts_multi_nroots=None,
                 cc_state_to_opt=None,
                 cis_state_deriv=None,
                 RPA=False,
                 set_iter=30,
                 gui=0,
                 # optimization
                 geom_opt_coords=-1,
                 geom_opt_tol_gradient=300,
                 geom_opt_tol_displacement=1200,
                 geom_opt_tol_energy=100,
                 geom_opt_max_cycles=50,
                 # other
                 n_frozen_core=None,
                 n_frozen_virt=None,
                 namd_nsurfaces=None,
                 scf_print=None,
                 scf_guess=None,
                 mem_total=2000,
                 mem_static=64
                 ):

        # put to arguments self._* (will be written explicitly)
        for name, value in vars().items():
            if name != 'self':
                # set keywords in lower case
                if type(value) is str:
                    value = value.lower()
                setattr(self, '_' + name, value)

        # set ras_occ
        if correlation is not None:
            if ras_occ is None and correlation.upper() == 'RASCI':
                if ras_elec is not None:
                    self._ras_occ = (np.sum(
                        molecule.get_atomic_numbers()) - ras_elec - molecule.charge) // 2
                elif ras_elec_alpha is not None or ras_elec_beta is not None:
                    self._ras_occ = (np.sum(
                        molecule.get_atomic_numbers()) - ras_elec_alpha - ras_elec_beta - molecule.charge) // 2
                else:
                    self._ras_occ = (np.sum(molecule.get_atomic_numbers()) - molecule.charge) // 2
                self._ras_occ = int(self._ras_occ)
                warnings.warn(QchemInputWarning('set ras_occ = {}'.format(self._ras_occ)))

        # Handle custom basis set
        if type(basis) is not str:
            self._basis = 'gen'
            self._custom_basis = basis


        # handle explicit guess (from MO coefficients)
        if scf_guess is not None and type(scf_guess) is not str:
            # print('set custom guess')
            self._scf_guess = 'read'
            self._mo_coefficients = scf_guess
        else:
            self._mo_coefficients = None

        if self._ras_srdft is not None:
            from warnings import warn
            warn('Warning! ras_srdft keyword is deprecated, this will be automatically '
                 'activated when using ras_srdft_exc and ras_srdft_cor')
        if ras_srdft_exc is not None or ras_srdft_cor is not None:
           self._ras_srdft = True
        else:
            self._ras_srdft = False

    def __hash__(self):

        # take all keywords defined in input
        keywords = dict(self.__dict__)

        # remove keywords that not affect the results
        for key in ['_mem_total', '_mem_static', '_gui', '_set_iter']:
            keywords.pop(key, None)

        # Change molecule object by molecule coordinates (Structure class too complex for JSON)
        keywords['_molecule'] = hash(keywords['_molecule'])

        digest = hashlib.md5(json.dumps(keywords, sort_keys=True).encode()).hexdigest()
        return int(digest, 16)

    def get_txt(self):
        """
        get qchem input in plain text

        :return string: qchem input in plain text
        """

        input_file = ''

        # Molecule definition
        input_file += '$molecule\n'

        input_file += '{} {}\n'.format(self._molecule.charge, self._molecule.multiplicity)

        atomic_elements = self._molecule.get_atomic_elements()

        coordinates = self._molecule.get_coordinates()

        for index, element in enumerate(atomic_elements):
            input_file += (element + '\t' + '{:20.10f} {:20.10f} {:20.10f}\n'.format(*coordinates[index]))

        input_file += '$end\n'

        # Rem variables
        input_file += '$rem\n'
        input_file += 'jobtype {}\n'.format(self._jobtype)

        if self._exchange is not None:
            input_file += 'exchange {}\n'.format(self._exchange)
        else:
            input_file += 'method {}\n'.format(self._method)

        input_file += 'basis {}\n'.format(self._basis)
        input_file += 'thresh {}\n'.format(self._thresh)
        input_file += 'scf_convergence {}\n'.format(self._scf_convergence)
        input_file += 'max_scf_cycles {}\n'.format(self._max_scf_cycles)
        input_file += 'gui {}\n'.format(self.gui)
        input_file += 'set_iter {}\n'.format(self._set_iter)
        input_file += 'RPA {}\n'.format(self._RPA)
        input_file += 'mem_total {}\n'.format(self._mem_total)
        input_file += 'mem_static {}\n'.format(self._mem_static)

        if self._unrestricted is not None:
            input_file += 'unrestricted {}\n'.format(self._unrestricted)

        if self._purecart is not None:
            input_file += 'purecart {}\n'.format(self._purecart)

        if self._correlation is not None:
            input_file += 'correlation {}\n'.format(self._correlation)

            # RasCI variables
            if self._correlation.upper() == 'RASCI':

                if self._ras_natorb_state is not None:
                    input_file += 'ras_natorb_state {}\n'.format(self._ras_natorb_state)
                    self._ras_natorb = True

                input_file += 'ras_roots {}\n'.format(self._ras_roots)
                input_file += 'ras_do_hole {}\n'.format(self._ras_do_hole)
                input_file += 'ras_do_part {}\n'.format(self._ras_do_part)
                input_file += 'ras_occ {}\n'.format(self._ras_occ)
                input_file += 'ras_spin_mult {}\n'.format(self._ras_spin_mult)
                input_file += 'ras_print {}\n'.format(self._ras_print)
                input_file += 'ras_natorb {}\n'.format(self._ras_natorb)
                input_file += 'ras_sts_tm {}\n'.format(self._ras_sts_tm)
                # input_file += 'max_cis_cycles {}\n'.format(self._max_cis_cycles)
                # input_file += 'RAS_RESTR_TYPE {}\n'.format(True)

                if self._ras_act is not None:
                    input_file += 'ras_act {}\n'.format(self._ras_act)
                else:
                    print('test')
                    raise Exception('{} not defined'.format('ras_act'))

                if self._ras_elec is not None:
                    input_file += 'ras_elec {}\n'.format(self._ras_elec)
                else:
                    if self._ras_elec_alpha is None and self._ras_elec_beta is None:
                        raise Exception('{} not defined'.format('ras_elec'))

                if self._ras_elec_alpha is not None:
                    input_file += 'ras_elec_alpha {}\n'.format(self._ras_elec_alpha)
                if self._ras_elec_beta is not None:
                    input_file += 'ras_elec_beta {}\n'.format(self._ras_elec_beta)

                if self._ras_act is not None:
                    input_file += 'ras_act {}\n'.format(self._ras_act)
                else:
                    raise Exception('{} not defined'.format('ras_act'))

                # Sr-DFT
                if self._ras_srdft:
                    input_file += 'ras_srdft {}\n'.format('True')
                    input_file += 'ras_srdft_damp {}\n'.format(self._ras_srdft_damp)
                    input_file += 'ras_srdft_spinpol {}\n'.format(self._ras_srdft_spinpol)
                    input_file += 'ras_omega {}\n'.format(self._ras_omega)

                    if self._ras_srdft_exc is not None:
                        input_file += 'ras_srdft_exc {}\n'.format(self._ras_srdft_exc)
                    else:
                        raise Exception('{} not defined'.format('ras_srdft_exc'))

                    if self._ras_srdft_cor is not None:
                        input_file += 'ras_srdft_cor {}\n'.format(self._ras_srdft_cor)
                    else:
                        raise Exception('{} not defined'.format('ras_srdft_cor'))

                # Diabatization
                diab_methods = {'ER': 1, 'Boys': 2, 'DQ': 3}
                if self._ras_diabatization_states is not None:
                    input_file += 'sts_multi_nroots {}\n'.format(len(self._ras_diabatization_states))
                    input_file += 'cis_diabath_decompose {}\n'.format(len(self._ras_diabatization_scheme))

                    input_file += 'ras_diab_seq_data ['
                    for seq in self._ras_diabatization_scheme:
                        input_file += '{} '.format([num for num in seq['states']] +
                                                   [diab_methods[seq['method']]] +
                                                   [seq['parameters'] if 'parameters' in seq else 0.0]).replace(' ', '')[1:-1]
                        input_file += ','
                    input_file = input_file[:-1] + ']\n'
                    input_file += 'ras_diab_seq_list ' + '{}\n'.format([len(seq['states']) for seq in self._ras_diabatization_scheme]).replace(' ', '')
        # SOC
        if self._calc_soc is not False:
            input_file += 'calc_soc {}\n'.format(self._calc_soc)

        # CIS variables
        if self._cis_n_roots is not None:
            input_file += 'cis_convergence {}\n'.format(self._cis_convergence)
            input_file += 'cis_n_roots {}\n'.format(self._cis_n_roots)
            input_file += 'cis_singlets {}\n'.format(self._cis_singlets)
            input_file += 'cis_triplets {}\n'.format(self._cis_triplets)
            input_file += 'cis_ampl_anal {}\n'.format(self._cis_ampl_anal)
            input_file += 'loc_cis_ov_separate {}\n'.format(self._loc_cis_ov_separate)
            input_file += 'er_cis_numstate {}\n'.format(self._er_cis_numstate)
            input_file += 'cis_diabath_decompose {}\n'.format(self._cis_diabath_decompose)
            input_file += 'max_cis_cycles {}\n'.format(self._max_cis_cycles)
        # other
        if self._n_frozen_core is not None:
            input_file += 'n_frozen_core {}\n'.format(self._n_frozen_core)
        if self._n_frozen_virt is not None:
            input_file += 'n_frozen_virtual {}\n'.format(self._n_frozen_virt)
        if self._namd_nsurfaces is not None:
            input_file += 'namd_nsurfaces {}\n'.format(self._namd_nsurfaces)
        if self._sts_multi_nroots is not None:
            input_file += 'sts_multi_nroots {}\n'.format(self._sts_multi_nroots)
        if self._localized_diabatization is not None:
            input_file += 'cis_diabath_decompose {}\n'.format(self._cis_diabath_decompose)
        if self._cc_state_to_opt is not None:
            input_file += 'cc_state_to_opt [{},{}]\n'.format(self._cc_state_to_opt[0], self._cc_state_to_opt[1])
        if self._cis_state_deriv is not None:
            input_file += 'cis_state_deriv {}\n'.format(self._cis_state_deriv)

        if self._scf_print is not None:
            input_file += 'scf_print {}\n'.format(self._scf_print)

        if self._scf_guess is not None:
            input_file += 'scf_guess {}\n'.format(self._scf_guess)

        input_file += '$end\n'

        # localized diabatization
        if self._localized_diabatization is not None:
            input_file += '$localized_diabatization\nadiabatic states\n'
            input_file += ' '.join(np.array(self._localized_diabatization, dtype=str))
            input_file += '\n$end\n'

        # optimization
        if self._jobtype.lower() == 'opt':
            input_file += 'geom_opt_coords {}\n'.format(self._geom_opt_coords)
            input_file += 'geom_opt_tol_gradient {}\n'.format(self._geom_opt_tol_gradient)
            input_file += 'geom_opt_tol_displacement {}\n'.format(self._geom_opt_tol_displacement)
            input_file += 'geom_opt_tol_energy {}\n'.format(self._geom_opt_tol_energy)
            input_file += 'geom_opt_max_cycles {}\n'.format(self._geom_opt_max_cycles)

        # Diabatization section
        if self._ras_diabatization_states is not None:
            input_file += '$localized_diabatization\n'
            input_file += 'adiabatic states\n'
            input_file += ' '.join([str(num) for num in self._ras_diabatization_states])
            input_file += '\n$end\n'

        # custom basis section
        if self._basis == 'gen':
            input_file += '$basis\n'
            input_file += basis_to_txt(self._custom_basis)
            input_file += '$end\n'

        return input_file + "\n"

    # Access to properties (only a reduced set should be accessible/editable)
    @property
    def mo_coefficients(self):
        return self._mo_coefficients

    @property
    def gui(self):
        return self._gui

    @gui.setter
    def gui(self, value):
        value = int(value)
        if value < 0 or value > 10:
            raise ValueError('GUI value error')
        self._gui = value

    def get_copy(self):
        """
        Get a copy of the input

        :return:
        """
        return deepcopy(self)

    def update_input(self, dictionary):
        """
        Update the input from data in a dictionary
        Note: already existing parameters will be overwritten

        :param dictionary: parameters to add
        """
        # put to arguments self._* (will be written explicitly)
        for name, value in dictionary.items():
            setattr(self, '_' + name, value)



__author__ = 'Abel Carreras'
import tempfile
import os
import numpy as np
import hashlib, json

def int_to_xyz(molecule, no_dummy=True):

    internal = molecule._get_full_z_matrix()
    coordinates = [[0.0, 0.0, 0.0]]

    for line in internal[1:]:
        bi = int(line[0])  # bond index
        B = line[1]        # bond value
        ai = int(line[2])  # Angle index
        A = line[3]        # Angle value
        ci = int(line[4])  # Dihedral index
        C = line[5]        # Dihedral value

        bond = np.array(coordinates[ai-1]) - np.array(coordinates[bi-1])
        if np.linalg.norm(bond) == 0:
            bond = np.array([1, 0, 0])

        bond2 = np.array(coordinates[ci-1]) - np.array(coordinates[ai-1])
        if np.linalg.norm(bond2) == 0:
            bond2 = np.array([0, 1, 0])

        origin = bond/np.linalg.norm(bond)*B
        ref2 = bond
        ref3 = np.cross(bond, bond2)

        # Check case of linear structure
        if np.linalg.norm(ref3) == 0:
            ref3 = [0.0, 0.0, 0.1]

        inter = np.dot(rotation_matrix(ref3, np.deg2rad(A)), origin)
        final = np.dot(rotation_matrix(ref2, np.deg2rad(C)), inter)
        final = final + np.array(coordinates[bi-1])
        coordinates.append(final)

    coordinates = np.array(coordinates)

    if no_dummy:
      #  mask = np.argwhere(molecule.get_atomic_elements_with_dummy()[:,0]  == 'X')
        mask = np.argwhere((molecule.get_atomic_elements_with_dummy()[:,0] == 'X') |
                           (molecule.get_atomic_elements_with_dummy()[:,0] == 'x')).flatten()
        coordinates = np.delete(coordinates,mask,axis=0)

    return np.array(coordinates, dtype=float)


def rotation_matrix(axis, theta):
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.
    """
    if np.dot(axis, axis) == 0.0:
        print ('Warning, reference rotation axis module is 0')
        exit()

    axis = np.asarray(axis)
    theta = np.asarray(theta)
    axis = axis/np.sqrt(np.dot(axis, axis))

    a = np.cos(theta/2)
    b, c, d = -axis*np.sin(theta/2)
    aa, bb, cc, dd = a*a, b*b, c*c, d*d
    bc, ad, ac, ab, bd, cd = b*c, a*d, a*c, a*b, b*d, c*d
    return np.array([[aa+bb-cc-dd, 2*(bc+ad), 2*(bd-ac)],
                     [2*(bc-ad), aa+cc-bb-dd, 2*(cd+ab)],
                     [2*(bd+ac), 2*(cd-ab), aa+dd-bb-cc]])


class Structure:

    def __init__(self,
                 coordinates=None,
                 internal=None,
                 z_matrix=None,
                 int_label=None,
                 atom_types=None,
                 atomic_elements=None,
                 atomic_numbers=None,
                 connectivity=None,
                 file_name=None,
                 charge=0,
                 multiplicity=1,
                 name=None,

                 #Buscar un lloc millor
                 int_weights=None):

        self._coordinates = np.array(coordinates)
        self._internal = internal
        self._z_matrix = z_matrix
        self._int_label = int_label
        self._atom_types = atom_types
        self._atomic_numbers = atomic_numbers
        self._connectivity = connectivity
        self._atomic_elements = atomic_elements
        self._charge = charge
        self._multiplicity = multiplicity
        self._name = name

        self._file_name = file_name
        self._int_weights = int_weights

        self._atomic_masses = None
        self._number_of_atoms = None
        self._number_of_internal = None
        self._energy = {}
        self._modes = None

        self._full_z_matrix = None

        # get electrons

    def __str__(self):
        return self.get_xyz()

    def __hash__(self):
        digest = hashlib.md5(json.dumps((self.get_xyz(), self.alpha_electrons, self.beta_electrons),
                                        sort_keys=True).encode()).hexdigest()
        return int(digest, 16)

    def get_coordinates(self):
        if self._coordinates is None:
            self._coordinates = int_to_xyz(self)
        return list(self._coordinates)

    def set_coordinates(self, coordinates):
        self._coordinates = np.array(coordinates)
        self._number_of_atoms = None
        self._energy = {}

    def _get_internal(self):
        if self._internal is None:
            print('No internal coordinates available\n Load internal file')
            exit()
        return self._internal.copy()

    def _set_internal(self, internal):
        self._internal = internal
        self._energy = None
        self._coordinates = int_to_xyz(self)
        self._full_z_matrix = None

    def _get_full_z_matrix(self):
        if self._full_z_matrix is None:
            num_z_atoms = self._get_z_matrix().shape[0]
            self._full_z_matrix = np.zeros((num_z_atoms,6))

            for row, i in enumerate(self._get_z_matrix()[1:]):
                    for col, k in enumerate(i[0]):
                        try:
                            self._full_z_matrix[row+1, col] = float(k)
                        except ValueError:
                            self._full_z_matrix[row+1, col] = self._get_int_dict()[k]


        return self._full_z_matrix

    def _get_z_matrix(self):
        if self._z_matrix is None:
            print('No Z-matrix available\n Load zmatrix file')
            exit()
        return self._z_matrix

    def _set_z_matrix(self, z_matrix):
        self._z_matrix = z_matrix

    def _get_int_label(self):
        return self._int_label

    def _set_int_label(self, int_label):
        self._int_label = int_label

    def _get_int_dict(self):
        self._internal_dict = {}
        for i, label in enumerate(self._get_int_label()[:, 0]):
            self._internal_dict.update({label:self._get_internal()[i, 0]})
        return self._internal_dict

    def _get_int_weights(self):
        return self._int_weights

    def _set_int_weights(self, int_weights):
        self._int_weights = int_weights

    def get_atomic_elements_with_dummy(self):
       # print([i for i in self._atomic_elements if i != "X"])
       return self._atomic_elements

    @property
    def name(self):
        return self._name

    @property
    def file_name(self):
        return self._file_name

    @file_name.setter
    def file_name(self, file_name):
        self._file_name = file_name

    @property
    def charge(self):
        return self._charge

    @charge.setter
    def charge(self, charge):
        self._charge = charge

    @property
    def multiplicity(self):
        return self._multiplicity

    @multiplicity.setter
    def multiplicity(self, multiplicity):
        self._multiplicity = multiplicity

    @property
    def number_of_electrons(self):
        return int(np.sum(self.get_atomic_numbers()) + self.charge)

    @property
    def alpha_electrons(self):
        alpha_unpaired = self.multiplicity // 2 + 1 if (self.number_of_electrons % 2) else self.multiplicity // 2
        return self.number_of_electrons // 2 + alpha_unpaired

    @property
    def beta_electrons(self):
        return self.number_of_electrons - self.alpha_electrons

    def _get_atom_types(self):
        if self._atom_types is None:
            print('No atom types available')
            exit()
        return self._atom_types

    def _set_atom_types(self, atom_types):
        self._atom_types = atom_types

    def get_atomic_numbers(self):
        if self._atomic_numbers is None:
            self._atomic_numbers = [[data[1].upper() for data in atom_data].index(element.upper())
                                    for element in self.get_atomic_elements()]
        return self._atomic_numbers

    def set_atomic_numbers(self, atomic_numbers):
        self._atomic_numbers = atomic_numbers

    def get_atomic_elements(self):
        if self._atomic_elements is None:
            self._atomic_elements = np.array(atom_data)[self.get_atomic_numbers()].T[1]
        return np.array([i for i in self._atomic_elements if i != "X"], dtype=str)

    def set_atomic_elements(self, atomic_elements):
        self._atomic_elements = atomic_elements

    def _get_connectivity(self):
        if self._connectivity is None:
            print('No atom connectivity available')
            exit()

        return self._connectivity

    def _set_connectivity(self, connectivity):
        self._connectivity = connectivity

#   Real methods

    def get_number_of_atoms(self):
        if self._number_of_atoms is None:
            self._number_of_atoms = np.array(self.get_coordinates()).shape[0]

        return self._number_of_atoms

    def _get_number_of_internal(self):
        if self._number_of_internal is None:
            self._number_of_internal = self._get_internal().shape[0]

        return self._number_of_internal

    def _get_energy(self, method=None):
        if method is None:
            if len(self._energy) == 1:
                return self._energy.values()[0]
            raise Exception('No method defined')
        elif '{}'.format(hash(method)) not in self._energy:
            self._energy['{}'.format(hash(method))] = method.single_point(self)
        return self._energy['{}'.format(hash(method))]

    def _get_modes(self, method=None):
        if self._modes is None:
            if method is None:
                raise Exception('No method defined')
            self._modes, energy = method.vibrations(self)
            self._energy['{}'.format(method.multiplicity)] = energy
        return self._modes

    def get_atomic_masses(self):
        if self._atomic_masses is None:

            try:
                masses_string = np.array(atom_data)[:, 3:4][[np.where(np.array(atom_data)==element)[0][0]
                                                             for element in self.get_atomic_elements()]]
                self._atomic_masses = np.array(masses_string, dtype=float).T[0]
            except TypeError:
                print('Error reading element labels')
                exit()
        return self._atomic_masses

    def get_valence_electrons(self):
        valence_electrons = 0
        for number in self.get_atomic_numbers():
            if 2 >= number > 0:
                valence_electrons += np.mod(number, 2)
            if 18 >= number > 2:
                valence_electrons += np.mod(number-2, 8)
            if 54 >= number > 18:
                valence_electrons += np.mod(number-18, 18)
            if 118 >= number > 54:
                valence_electrons += np.mod(number-54, 32)
            if number > 118:
                raise Exception('Atomic number size not implemented')

        valence_electrons -= self.charge

        return valence_electrons

    def get_xyz(self, title=''):
        txt = '{}\n{}\n'.format(self.get_number_of_atoms(), title)
        for s, c in zip(self.get_atomic_elements(), self.get_coordinates()):
            txt += '{:2} '.format(s) + '{:15.10f} {:15.10f} {:15.10f}\n'.format(*c)

        return txt


atom_data = [
    [  0, "X", "X", 0], # 0
    [  1, "H", "Hydrogen", 1.00794], # 1
    [  2, "He", "Helium", 4.002602], # 2
    [  3, "Li", "Lithium", 6.941], # 3
    [  4, "Be", "Beryllium", 9.012182], # 4
    [  5, "B", "Boron", 10.811], # 5
    [  6, "C", "Carbon", 12.0107], # 6
    [  7, "N", "Nitrogen", 14.0067], # 7
    [  8, "O", "Oxygen", 15.9994], # 8
    [  9, "F", "Fluorine", 18.9984032], # 9
    [ 10, "Ne", "Neon", 20.1797], # 10
    [ 11, "Na", "Sodium", 22.98976928], # 11
    [ 12, "Mg", "Magnesium", 24.3050], # 12
    [ 13, "Al", "Aluminium", 26.9815386], # 13
    [ 14, "Si", "Silicon", 28.0855], # 14
    [ 15, "P", "Phosphorus", 30.973762], # 15
    [ 16, "S", "Sulfur", 32.065], # 16
    [ 17, "Cl", "Chlorine", 35.453], # 17
    [ 18, "Ar", "Argon", 39.948], # 18
    [ 19, "K", "Potassium", 39.0983], # 19
    [ 20, "Ca", "Calcium", 40.078], # 20
    [ 21, "Sc", "Scandium", 44.955912], # 21
    [ 22, "Ti", "Titanium", 47.867], # 22
    [ 23, "V", "Vanadium", 50.9415], # 23
    [ 24, "Cr", "Chromium", 51.9961], # 24
    [ 25, "Mn", "Manganese", 54.938045], # 25
    [ 26, "Fe", "Iron", 55.845], # 26
    [ 27, "Co", "Cobalt", 58.933195], # 27
    [ 28, "Ni", "Nickel", 58.6934], # 28
    [ 29, "Cu", "Copper", 63.546], # 29
    [ 30, "Zn", "Zinc", 65.38], # 30
    [ 31, "Ga", "Gallium", 69.723], # 31
    [ 32, "Ge", "Germanium", 72.64], # 32
    [ 33, "As", "Arsenic", 74.92160], # 33
    [ 34, "Se", "Selenium", 78.96], # 34
    [ 35, "Br", "Bromine", 79.904], # 35
    [ 36, "Kr", "Krypton", 83.798], # 36
    [ 37, "Rb", "Rubidium", 85.4678], # 37
    [ 38, "Sr", "Strontium", 87.62], # 38
    [ 39, "Y", "Yttrium", 88.90585], # 39
    [ 40, "Zr", "Zirconium", 91.224], # 40
    [ 41, "Nb", "Niobium", 92.90638], # 41
    [ 42, "Mo", "Molybdenum", 95.96], # 42
    [ 43, "Tc", "Technetium", 0], # 43
    [ 44, "Ru", "Ruthenium", 101.07], # 44
    [ 45, "Rh", "Rhodium", 102.90550], # 45
    [ 46, "Pd", "Palladium", 106.42], # 46
    [ 47, "Ag", "Silver", 107.8682], # 47
    [ 48, "Cd", "Cadmium", 112.411], # 48
    [ 49, "In", "Indium", 114.818], # 49
    [ 50, "Sn", "Tin", 118.710], # 50
    [ 51, "Sb", "Antimony", 121.760], # 51
    [ 52, "Te", "Tellurium", 127.60], # 52
    [ 53, "I", "Iodine", 126.90447], # 53
    [ 54, "Xe", "Xenon", 131.293], # 54
    [ 55, "Cs", "Caesium", 132.9054519], # 55
    [ 56, "Ba", "Barium", 137.327], # 56
    [ 57, "La", "Lanthanum", 138.90547], # 57
    [ 58, "Ce", "Cerium", 140.116], # 58
    [ 59, "Pr", "Praseodymium", 140.90765], # 59
    [ 60, "Nd", "Neodymium", 144.242], # 60
    [ 61, "Pm", "Promethium", 0], # 61
    [ 62, "Sm", "Samarium", 150.36], # 62
    [ 63, "Eu", "Europium", 151.964], # 63
    [ 64, "Gd", "Gadolinium", 157.25], # 64
    [ 65, "Tb", "Terbium", 158.92535], # 65
    [ 66, "Dy", "Dysprosium", 162.500], # 66
    [ 67, "Ho", "Holmium", 164.93032], # 67
    [ 68, "Er", "Erbium", 167.259], # 68
    [ 69, "Tm", "Thulium", 168.93421], # 69
    [ 70, "Yb", "Ytterbium", 173.054], # 70
    [ 71, "Lu", "Lutetium", 174.9668], # 71
    [ 72, "Hf", "Hafnium", 178.49], # 72
    [ 73, "Ta", "Tantalum", 180.94788], # 73
    [ 74, "W", "Tungsten", 183.84], # 74
    [ 75, "Re", "Rhenium", 186.207], # 75
    [ 76, "Os", "Osmium", 190.23], # 76
    [ 77, "Ir", "Iridium", 192.217], # 77
    [ 78, "Pt", "Platinum", 195.084], # 78
    [ 79, "Au", "Gold", 196.966569], # 79
    [ 80, "Hg", "Mercury", 200.59], # 80
    [ 81, "Tl", "Thallium", 204.3833], # 81
    [ 82, "Pb", "Lead", 207.2], # 82
    [ 83, "Bi", "Bismuth", 208.98040], # 83
    [ 84, "Po", "Polonium", 0], # 84
    [ 85, "At", "Astatine", 0], # 85
    [ 86, "Rn", "Radon", 0], # 86
    [ 87, "Fr", "Francium", 0], # 87
    [ 88, "Ra", "Radium", 0], # 88
    [ 89, "Ac", "Actinium", 0], # 89
    [ 90, "Th", "Thorium", 232.03806], # 90
    [ 91, "Pa", "Protactinium", 231.03588], # 91
    [ 92, "U", "Uranium", 238.02891], # 92
    [ 93, "Np", "Neptunium", 0], # 93
    [ 94, "Pu", "Plutonium", 0], # 94
    [ 95, "Am", "Americium", 0], # 95
    [ 96, "Cm", "Curium", 0], # 96
    [ 97, "Bk", "Berkelium", 0], # 97
    [ 98, "Cf", "Californium", 0], # 98
    [ 99, "Es", "Einsteinium", 0], # 99
    [100, "Fm", "Fermium", 0], # 100
    [101, "Md", "Mendelevium", 0], # 101
    [102, "No", "Nobelium", 0], # 102
    [103, "Lr", "Lawrencium", 0], # 103
    [104, "Rf", "Rutherfordium", 0], # 104
    [105, "Db", "Dubnium", 0], # 105
    [106, "Sg", "Seaborgium", 0], # 106
    [107, "Bh", "Bohrium", 0], # 107
    [108, "Hs", "Hassium", 0], # 108
    [109, "Mt", "Meitnerium", 0], # 109
    [110, "Ds", "Darmstadtium", 0], # 110
    [111, "Rg", "Roentgenium", 0], # 111
    [112, "Cn", "Copernicium", 0], # 112
    [113, "Uut", "Ununtrium", 0], # 113
    [114, "Uuq", "Ununquadium", 0], # 114
    [115, "Uup", "Ununpentium", 0], # 115
    [116, "Uuh", "Ununhexium", 0], # 116
    [117, "Uus", "Ununseptium", 0], # 117
    [118, "Uuo", "Ununoctium", 0], # 118
    ]



from setuptools import setup

def get_version_number():
    for l in open('pyqchem/__init__.py', 'r').readlines():
        if not(l.find('__version__')):
            exec(l, globals())
            return __version__


# Make python package
setup(name='pyqchem',
      version=get_version_number(),
      description='Python wrapper for Q-Chem',
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      install_requires=['numpy', 'scipy', 'lxml', 'requests', 'matplotlib', 'PyYAML'],
      author='Abel Carreras',
      author_email='abelcarreras83@gmail.com',
      packages=['pyqchem', 'pyqchem.parsers'],
      url='https://github.com/abelcarreras/PyQchem',
      classifiers=[
          "Programming Language :: Python",
          "License :: OSI Approved :: MIT License"]
      )



import numpy as np
from pyqchem.structure import Structure


def basis_format(basis_set_name,
                 atomic_numbers,
                 atomic_symbols,
                 shell_type,
                 n_primitives,
                 atom_map,
                 p_exponents,
                 c_coefficients,
                 p_c_coefficients):
    """
    Function to generate standard basis dictionary

    :param basis_set_name:
    :param atomic_numbers:
    :param atomic_symbols:
    :param shell_type:
    :param n_primitives:
    :param atom_map:
    :param p_exponents:
    :param c_coefficients:
    :param p_c_coefficients:
    :return:
    """

    # print(n_primitives)

    typeList = {'0': ['s', 1],
                '1': ['p', 3],
                '2': ['d', 6],
                '3': ['f', 10],
                '-1': ['sp', 4],
                '-2': ['d', 5],
                '-3': ['f', 7]}

    atomic_numbers = [int(an) for an in atomic_numbers]
    atom_map = np.array(atom_map, dtype=int)
    # print(atom_map)
    basis_set = {'name': basis_set_name,
                 'primitive_type': 'gaussian'}

    shell_type_index = [0] + np.cumsum([typeList['{}'.format(s)][1]
                                        for s in shell_type]).tolist()
    prim_from_shell_index = [0] + np.cumsum(np.array(n_primitives, dtype=int)).tolist()

    # print(shell_type_index)
    # print(prim_from_shell_index)

    atoms_data = []
    for iatom, atomic_number in enumerate(atomic_numbers):
        symbol = atomic_symbols[iatom]

        shell_from_atom_counts = np.unique(atom_map, return_counts=True)[1]
        shell_from_atom_index = np.unique(atom_map, return_index=True)[1]
        # print(shell_from_atom_counts)
        # print('atom_indexes', shell_from_atom_index)
        # print('atom_number', iatom)
        # print('shells index', shell_from_atom_index[iatom])
        # print('number of shells', shell_from_atom_counts[iatom])

        shells_data = []
        for ishell in range(shell_from_atom_counts[iatom]):
            st = typeList['{}'.format(shell_type[shell_from_atom_index[iatom] + ishell])]
            # print(st, ishell)
            ini_prim = prim_from_shell_index[shell_from_atom_index[iatom] + ishell]
            fin_prim = prim_from_shell_index[shell_from_atom_index[iatom] + ishell+1]
            # print(ini_prim)
            # print(fin_prim)

            shells_data.append({
                'shell_type': st[0],
                'functions': st[1],
                'p_exponents': p_exponents[ini_prim: fin_prim],
                'con_coefficients': c_coefficients[ini_prim: fin_prim],
                'p_con_coefficients': p_c_coefficients[ini_prim: fin_prim],
            })

        atoms_data.append({'shells': shells_data,
                           'symbol': symbol,
                           'atomic_number': atomic_number})

    basis_set['atoms'] = atoms_data

    return basis_set


def _reformat_input(array):
    flat_list = []
    for sublist in array:
        for item in sublist:
            if len(item) > 2:
                flat_list.append(item)
            else:
                flat_list.append(item)
    return flat_list


def vect_to_mat(vector):
    n = int(np.sqrt(0.25 + 2 * len(vector)) - 0.5)

    k = 0
    matrix = np.zeros([n, n])
    for i in range(n):
        for j in range(0, i+1):
            matrix[i, j] = vector[k + j]
            matrix[j, i] = vector[k + j]
        k += i+1

    return matrix


def _get_all_nato(output):
    import re

    nato_coefficients_list = []
    nato_occupancies_list = []

    for m in re.finditer('Alpha NATO coefficients', output):
        n_elements = int(output[m.end():m.end() + 100].replace('\n', ' ').split()[2])
        nbas = int(np.sqrt(n_elements))
        data = output[m.end(): m.end() + n_elements*50].split()[3:n_elements+3]
        nato_coefficients_list.append({'alpha': np.array(data, dtype=float).reshape(nbas, nbas).tolist()})

    for m in re.finditer('Alpha Natural Orbital occupancies', output):
        n_elements = int(output[m.end():m.end() + 100].replace('\n', ' ').split()[2])
        data = output[m.end(): m.end() + n_elements*50].split()[3:n_elements+3]
        nato_occupancies_list.append({'alpha': np.array(data, dtype=float).tolist()})

    for i, m in enumerate(re.finditer('Beta NATO coefficients', output)):
        n_elements = int(output[m.end():m.end() + 100].replace('\n', ' ').split()[2])
        nbas = int(np.sqrt(n_elements))
        data = output[m.end(): m.end() + n_elements*50].split()[3:n_elements+3]
        nato_coefficients_list[i]['beta'] = np.array(data, dtype=float).reshape(nbas, nbas).tolist()

    for i, m in enumerate(re.finditer('Beta Natural Orbital occupancies', output)):
        n_elements = int(output[m.end():m.end() + 100].replace('\n', ' ').split()[2])
        data = output[m.end(): m.end() + n_elements*50].split()[3:n_elements+3]
        nato_occupancies_list[i]['beta'] = np.array(data, dtype=float).tolist()

    return nato_coefficients_list, nato_occupancies_list


def parser_fchk(output):

    def convert_to_type(item_type, item):
        item_types = {'I': int,
                      'R': float}

        if type(item) is list:
            return [item_types[item_type](e) for e in item]
        else:
            return item_types[item_type](item)

    key_list = ['Charge', 'Multiplicity', 'Number of alpha electrons', 'Number of beta electrons',
                'Atomic numbers', 'Current cartesian coordinates', 'Shell types',
                'Number of primitives per shell', 'Shell to atom map', 'Primitive exponents',
                'Contraction coefficients', 'P(S=P) Contraction coefficients', 'Alpha MO coefficients',
                'Beta MO coefficients', 'Coordinates of each shell', 'Overlap Matrix',
                'Core Hamiltonian Matrix', 'Alpha Orbital Energies', 'Beta Orbital Energies',
                'Total SCF Density', 'Alpha NATO coefficients', 'Beta NATO coefficients',
                'Alpha Natural Orbital occupancies', 'Beta Natural Orbital occupancies'
                ]

    basis_set = output.split('\n')[1].split()[-1]
    words_output = output.replace('\n', ' ').split()

    data = {}
    nw = len(words_output)
    for key in key_list:
        wc = len(key.split())
        for i in range(nw):
            word = ' '.join(words_output[i:i+wc])
            if word == key:
                item_type = words_output[i+wc]
                if words_output[i + wc + 1] == 'N=':
                    n_elements = int(words_output[i + wc + 2])
                    data[word] = convert_to_type(item_type, words_output[i + wc + 3: i + wc + n_elements + 3])
                else:
                    data[word] = convert_to_type(item_type, words_output[i + wc + 1])
                break

    bohr_to_angstrom = 0.529177249

    coordinates = np.array(data['Current cartesian coordinates']).reshape(-1, 3) * bohr_to_angstrom
    structure = Structure(coordinates=coordinates,
                          atomic_numbers=data['Atomic numbers'],
                          multiplicity=data['Multiplicity'],
                          charge=data['Charge'])

    if not 'P(S=P) Contraction coefficients' in data:
        data['P(S=P) Contraction coefficients'] = np.zeros_like(data['Contraction coefficients']).tolist()

    basis = basis_format(basis_set_name=basis_set,
                         atomic_numbers=structure.get_atomic_numbers(),
                         atomic_symbols=structure.get_atomic_elements(),
                         shell_type=data['Shell types'],
                         n_primitives=data['Number of primitives per shell'],
                         atom_map=data['Shell to atom map'],
                         p_exponents=data['Primitive exponents'],
                         c_coefficients=data['Contraction coefficients'],
                         p_c_coefficients=data['P(S=P) Contraction coefficients'])

    nbas = int(np.sqrt(len(data['Alpha MO coefficients'])))
    mo_coeff = {'alpha': np.array(data['Alpha MO coefficients']).reshape(nbas, nbas).tolist()}
    mo_energy = {'alpha': data['Alpha Orbital Energies']}

    if 'Beta MO coefficients' in data:
        mo_coeff['beta'] = np.array(data['Beta MO coefficients']).reshape(nbas, nbas).tolist()
        mo_energy['beta'] = data['Beta Orbital Energies']

    final_dict = {'structure': structure,
                  'basis': basis,
                  'coefficients': mo_coeff,
                  'mo_energies': mo_energy,
                  'number_of_electrons': {'alpha': data['Number of alpha electrons'],
                                          'beta': data['Number of beta electrons']}
                  }

    if 'Overlap Matrix' in data:
        final_dict['overlap'] = vect_to_mat(data['Overlap Matrix']).tolist()

    if 'Alpha NATO coefficients' in data:
        final_dict['nato_coefficients'] = {'alpha': np.array(data['Alpha NATO coefficients']).reshape(nbas, nbas).tolist()}
        final_dict['nato_occupancies'] = {'alpha': data['Alpha Natural Orbital occupancies']}

    if 'Beta NATO coefficients' in data:
        final_dict['nato_coefficients'].update({
            'beta': np.array(data['Beta NATO coefficients']).reshape(nbas, nbas).tolist()})
        final_dict['nato_occupancies'].update({'beta': data['Beta Natural Orbital occupancies']})

    # check multiple NATO (may be improved)
    if 'Alpha NATO coefficients' in data:
        nato_coefficients_list, nato_occupancies_list = _get_all_nato(output)
        if len(nato_occupancies_list) > 1:
            final_dict['nato_coefficients_multi'] = nato_coefficients_list
            final_dict['nato_occupancies_multi'] = nato_occupancies_list

    return final_dict



__author__ = 'Abel Carreras'

import re
from pyqchem.utils import search_bars, standardize_vector


def basic_cis(output):
    """
    Parser for CIS calculations

    :param output:
    :return:
    """

    # scf energy
    enum = output.find('Total energy in the final basis set')
    scf_energy = float(output[enum:enum+100].split()[8])

    # CIS excited states
    # enum = output.find('CIS Excitation Energies')
    enum = list(re.finditer('CIS Excitation Energies', output))[-1].end()
    excited_states_cis = []
    if enum > 0:
        bars = search_bars(output, from_position=enum)

        output_cis = output[bars[0]:bars[1]]

        for m in re.finditer('Excited state ', output_cis):
            state_cis_words = output_cis[m.end():].split()

            exc_energy = float(state_cis_words[5])
            exc_energy_units = state_cis_words[3][1:-1]
            tot_energy = float(state_cis_words[11])
            try:
                tot_energy_units = state_cis_words[12]
                mul = state_cis_words[14]
                trans_mom = [float(mom) for mom in [state_cis_words[17],
                                                    state_cis_words[19],
                                                    state_cis_words[21]]]
                strength = float(state_cis_words[25])
            except ValueError:
                # old version of qchem (< 5.01)
                tot_energy_units = 'au'
                mul = state_cis_words[13]
                trans_mom = [float(mom) for mom in [state_cis_words[16],
                                                    state_cis_words[18],
                                                    state_cis_words[20]]]
                strength = float(state_cis_words[24])

            transitions = []
            for i, word in enumerate(state_cis_words):
                if word == '-->':
                    origin = int(state_cis_words[i-1][:-1])
                    target = int(state_cis_words[i+2][:-1])
                    amplitude = float(state_cis_words[i+5])

                    transitions.append({'origin': origin,
                                        'target': target,
                                        'amplitude': amplitude})
                if word == 'Excited':
                    break

            excited_states_cis.append({'total energy': tot_energy,
                                       'total energy units': tot_energy_units,
                                       'excitation energy': exc_energy,
                                       'excitation energy units': exc_energy_units,
                                       'multiplicity': mul,
                                       'transition moment': standardize_vector(trans_mom),
                                       'strength': strength,
                                       'transitions': transitions})

    return {'scf energy': scf_energy,
            'excited states cis': excited_states_cis}

